import datetime
from decimal import Decimal
from typing import List, Union, Tuple

import toloka.client as toloka

from crowdom import (
    base,
    evaluation,
    classification,
    classification_loop,
    duration,
    feedback_loop,
    control,
    mapping,
    pool as pool_config,
    worker,
)
from crowdom.objects import Audio, Text
from . import lib


class TolokaClientStub(lib.TolokaClientCallRecorderStub):
    def __init__(self, assignments: List[toloka.Assignment]):
        self.assignments = assignments
        super(TolokaClientStub, self).__init__()

    # will be called during checks retrieval
    def get_assignments(
        self,
        status: Union[toloka.Assignment.Status, List[toloka.Assignment.Status]],
        pool_id: str,
    ) -> List[toloka.Assignment]:
        if not isinstance(status, list):
            status = [status]
        return [assignment for assignment in self.assignments if assignment.status in status]

    # called by internal classification loop, mock closed pool
    def get_pool(self, pool_id: str) -> toloka.Pool:
        pool = toloka.Pool()
        pool.is_closed = lambda: True
        return pool

    # # called by internal classification loop, return empty because not check overlaps will change
    def get_tasks(self, request: toloka.search_requests.TaskSearchRequest) -> List[toloka.Task]:
        return []


def test_give_bonuses_to_users():
    audios = [Audio(url=f'https://storage.net/{i}.wav') for i in range(6)]

    markup_assignments = [
        # accuracy = 0.75
        lib.create_markup_assignment(
            audio_text_pairs=[
                # OK
                (audios[2], Text(text='алло')),
                # OK
                (audios[5], Text(text='привет')),
                # OK
                (audios[3], Text(text='четыре шесть')),
                # BAD
                (audios[0], Text(text='нет не надо')),
                # UNKNOWN
                (audios[4], Text(text='леруа ?')),
            ],
            id='m1',
            status=toloka.Assignment.ACCEPTED,
            user_id='a',
        ),
        # accuracy = 0.5
        lib.create_markup_assignment(
            audio_text_pairs=[
                # BAD
                (audios[0], Text(text='нет не надо')),
                # OK
                (audios[4], Text(text='леруа мерлен')),
                # BAD
                (audios[3], Text(text='четырнадцать')),
                # OK
                (audios[1], Text(text='хорошо')),
                # UNKNOWN
                (audios[5], Text(text='?')),
            ],
            id='m2',
            status=toloka.Assignment.REJECTED,
            user_id='c',
        ),
    ]

    def generate_check_assignment(id: str, data: List[Tuple[str, str, bool]], user_id: str) -> toloka.Assignment:
        return toloka.Assignment(
            id=id,
            user_id=user_id,
            tasks=[
                lib.audio_transcript_check_mapping.to_task((Audio(url=url), Text(text=text))) for url, text, _ in data
            ],
            solutions=[
                lib.audio_transcript_check_mapping.to_solution((base.BinaryEvaluation(ok=ok),)) for _, _, ok in data
            ],
            status=toloka.Assignment.ACCEPTED,
        )

    check_assignments = [
        generate_check_assignment(
            'c1',
            [
                ('https://storage.net/0.wav', 'нет не надо', False),
                ('https://storage.net/1.wav', 'хорошо', True),
                ('https://storage.net/2.wav', 'алло', True),
                ('https://storage.net/3.wav', 'четыре шесть', False),
                ('https://storage.net/4.wav', 'леруа мерлен', True),
                ('https://storage.net/5.wav', 'првет', False),
            ],
            'b',
        ),
        generate_check_assignment(
            'c2',
            [
                ('https://storage.net/0.wav', 'нет не надо', False),
                ('https://storage.net/1.wav', '', False),
                ('https://storage.net/2.wav', 'алло', True),
                ('https://storage.net/3.wav', 'четырнадцать', False),
                ('https://storage.net/4.wav', 'леруа мерлен', True),
                ('https://storage.net/5.wav', 'привет', True),
            ],
            'a',
        ),
        generate_check_assignment(
            'c3',
            [
                ('https://storage.net/0.wav', 'нет спасибо не надо', True),
                ('https://storage.net/1.wav', 'хорошо', False),
                ('https://storage.net/3.wav', 'четырнадцать', False),
                ('https://storage.net/3.wav', 'четыре шесть', True),
                ('https://storage.net/4.wav', '? ?', False),
                ('https://storage.net/5.wav', 'привет', False),
            ],
            'd',
        ),
    ]

    stub = TolokaClientStub(check_assignments)

    fb_loop = feedback_loop.FeedbackLoop(
        pool_input_objects=[(audio,) for audio in audios],
        markup_task_mapping=mapping.TaskMapping(input_mapping=(lib.audio_mapping,), output_mapping=(lib.text_mapping,)),
        check_task_mapping=mapping.TaskMapping(
            input_mapping=(lib.audio_mapping, lib.text_mapping), output_mapping=(lib.binary_evaluation_mapping,)
        ),
        check_params=classification_loop.Params(
            control=feedback_loop.Control(rules=control.RuleBuilder().add_static_reward(threshold=0.5).build()),
            overlap=classification_loop.StaticOverlap(overlap=2),
            aggregation_algorithm=classification.AggregationAlgorithm.MAJORITY_VOTE,
            task_duration_function=None,  # noqa
        ),
        markup_params=feedback_loop.Params(
            control=control.Control(
                rules=control.RuleBuilder()
                .add_dynamic_reward(
                    min_bonus_amount_usd=0.02,
                    max_bonus_amount_usd=0.04,
                    min_accuracy_for_bonus=0.5,
                    bonus_granularity_num=3,
                )
                .build()
            ),
            assignment_check_sample=None,
            overlap=classification_loop.DynamicOverlap(min_overlap=1, max_overlap=None, confidence=0.5),
            task_duration_function=None,
        ),
        client=stub,  # noqa
        lang='RU',
    )

    fb_loop.check_loop.pool_cfg = pool_config.ClassificationConfig(
        project_id=None,
        private_name=None,
        reward_per_assignment=None,
        task_duration_hint=None,  # noqa
        real_tasks_count=6,
        control_tasks_count=0,
    )

    fb_loop.give_bonuses_to_users(check_pool_id='fake', markup_assignments=markup_assignments)

    assert stub.calls == [
        (
            'create_user_bonuses_async',
            (
                [
                    toloka.user_bonus.UserBonus(
                        user_id='a', assignment_id='m1', amount=Decimal(0.03), without_message=True
                    ),
                    toloka.user_bonus.UserBonus(
                        user_id='c', assignment_id='m2', amount=Decimal(0.02), without_message=True
                    ),
                ],
            ),
        ),
        (
            'wait_operation',
            (toloka.operations.UserBonusCreateBatchOperation(status=toloka.operations.Operation.Status.SUCCESS),),
        ),
    ]


def test_get_results():
    # check all verdicts (BAD, UNKNOWN, OK)
    # check solutions sorting
    # - case then both solutions verdict is UNKNOWN and we prefer solution from better assignment

    assignment_id_to_accuracy_and_precision_recall = {
        '0': (1 / 3, 3 / 4),
        '1': (2 / 3, 3 / 4),
        '2': (1 / 2, 2 / 4),
        '3': (1 / 1, 1 / 2),
    }

    def generate_task_result(
        solutions: List[Tuple[str, feedback_loop.SolutionVerdict, str, str]],
    ) -> List[feedback_loop.Solution]:
        return [
            feedback_loop.Solution(
                solution=(Text(text=text),),
                verdict=verdict,
                worker=worker.Human(toloka.Assignment(user_id=user_id, id=assignment_id)),
                evaluation=evaluation.SolutionEvaluation(
                    ok=verdict == feedback_loop.SolutionVerdict.OK,
                    confidence=None,
                    worker_labels=[],  # noqa
                )
                if verdict != feedback_loop.SolutionVerdict.UNKNOWN
                else None,
                assignment_accuracy=assignment_id_to_accuracy_and_precision_recall[assignment_id][0],
                assignment_evaluation_recall=assignment_id_to_accuracy_and_precision_recall[assignment_id][1],
            )
            for text, verdict, assignment_id, user_id in solutions
        ]

    assert feedback_loop.get_results(
        pool_input_objects=[
            # test results order corresponds to pool input objects order
            (Audio(url=f'https://storage.net/{i}.wav'),)
            for i in range(6)
        ],
        markup_assignments=[
            lib.create_markup_assignment(
                audio_text_pairs=[
                    # UNKNOWN
                    (Audio(url='https://storage.net/2.wav'), Text(text='? до свидания')),
                    # BAD
                    (Audio(url='https://storage.net/5.wav'), Text(text='')),
                    # OK
                    (Audio(url='https://storage.net/0.wav'), Text(text='але')),
                    # BAD
                    (Audio(url='https://storage.net/1.wav'), Text(text='позови аператора')),
                ],
                id='0',
                status=toloka.Assignment.REJECTED,
                user_id='c',
            ),
            lib.create_markup_assignment(
                audio_text_pairs=[
                    # BAD
                    (Audio(url='https://storage.net/1.wav'), Text(text='позови аператора')),
                    # UNKNOWN
                    (Audio(url='https://storage.net/2.wav'), Text(text='да до свидания')),
                    # OK
                    (Audio(url='https://storage.net/3.wav'), Text(text='леруа мерлен')),
                    # OK
                    (Audio(url='https://storage.net/4.wav'), Text(text='нет не надо')),
                ],
                id='1',
                status=toloka.Assignment.REJECTED,
                user_id='a',
            ),
            lib.create_markup_assignment(
                audio_text_pairs=[
                    # UNKNOWN
                    (Audio(url='https://storage.net/3.wav'), Text(text='? мерлен')),
                    # BAD
                    (Audio(url='https://storage.net/2.wav'), Text(text='? досвидания')),
                    # OK
                    (Audio(url='https://storage.net/5.wav'), Text(text='нет')),
                    # UNKNOWN
                    (Audio(url='https://storage.net/1.wav'), Text(text='позовите оператора')),
                ],
                id='2',
                status=toloka.Assignment.REJECTED,
                user_id='b',
            ),
            lib.create_markup_assignment(
                audio_text_pairs=[
                    # OK
                    (Audio(url='https://storage.net/2.wav'), Text(text='да хорошо до свидания')),
                    # UNKNOWN
                    (Audio(url='https://storage.net/1.wav'), Text(text='угу позовите оператора')),
                ],
                id='3',
                status=toloka.Assignment.ACCEPTED,
                user_id='e',
            ),
        ],
        solution_id_to_evaluation=lib.generate_evaluations(
            [
                ('https://storage.net/5.wav', '', False),
                ('https://storage.net/0.wav', 'але', True),
                ('https://storage.net/1.wav', 'позови аператора', False),
                ('https://storage.net/3.wav', 'леруа мерлен', True),
                ('https://storage.net/4.wav', 'нет не надо', True),
                ('https://storage.net/2.wav', '? досвидания', False),
                ('https://storage.net/5.wav', 'нет', True),
                ('https://storage.net/2.wav', 'да хорошо до свидания', True),
            ]
        ),
        markup_task_mapping=lib.audio_transcript_mapping,
        check_task_mapping=lib.audio_transcript_check_mapping,
    ) == [
        generate_task_result(solutions=[('але', feedback_loop.SolutionVerdict.OK, '0', 'c')]),
        generate_task_result(
            solutions=[
                ('угу позовите оператора', feedback_loop.SolutionVerdict.UNKNOWN, '3', 'e'),
                ('позовите оператора', feedback_loop.SolutionVerdict.UNKNOWN, '2', 'b'),
                ('позови аператора', feedback_loop.SolutionVerdict.BAD, '1', 'a'),
                ('позови аператора', feedback_loop.SolutionVerdict.BAD, '0', 'c'),
            ],
        ),
        generate_task_result(
            solutions=[
                ('да хорошо до свидания', feedback_loop.SolutionVerdict.OK, '3', 'e'),
                ('да до свидания', feedback_loop.SolutionVerdict.UNKNOWN, '1', 'a'),
                ('? до свидания', feedback_loop.SolutionVerdict.UNKNOWN, '0', 'c'),
                ('? досвидания', feedback_loop.SolutionVerdict.BAD, '2', 'b'),
            ],
        ),
        generate_task_result(
            solutions=[
                ('леруа мерлен', feedback_loop.SolutionVerdict.OK, '1', 'a'),
                ('? мерлен', feedback_loop.SolutionVerdict.UNKNOWN, '2', 'b'),
            ],
        ),
        generate_task_result(
            solutions=[('нет не надо', feedback_loop.SolutionVerdict.OK, '1', 'a')],
        ),
        generate_task_result(
            solutions=[
                ('нет', feedback_loop.SolutionVerdict.OK, '2', 'b'),
                ('', feedback_loop.SolutionVerdict.BAD, '0', 'c'),
            ],
        ),
    ]


def test_speed_control():
    audios = [Audio(url=f'https://storage.net/{i}.wav') for i in range(6)]
    markup_assignment_solutions = [
        lib.create_markup_assignment(
            audio_text_pairs=[
                (audios[0], Text(text='алло')),
                (audios[1], Text(text='привет')),
            ],
            id='m1',
            user_id='a',
            duration=datetime.timedelta(seconds=40),
        ),
        lib.create_markup_assignment(
            audio_text_pairs=[
                (audios[2], Text(text='нет')),
                (audios[3], Text(text='да спасибо')),
            ],
            id='m2',
            user_id='b',
            duration=datetime.timedelta(seconds=10),
        ),
        lib.create_markup_assignment(
            audio_text_pairs=[
                (audios[4], Text(text='нет')),
                (audios[5], Text(text='да спасибо')),
            ],
            id='m3',
            user_id='c',
            duration=datetime.timedelta(seconds=3),
        ),
    ]
    stub = TolokaClientStub([assignment for assignment, _ in markup_assignment_solutions])

    fb_loop = feedback_loop.FeedbackLoop(
        pool_input_objects=[(audio,) for audio in audios],
        markup_task_mapping=mapping.TaskMapping(input_mapping=(lib.audio_mapping,), output_mapping=(lib.text_mapping,)),
        check_task_mapping=mapping.TaskMapping(
            input_mapping=(lib.audio_mapping, lib.text_mapping), output_mapping=(lib.binary_evaluation_mapping,)
        ),
        check_params=classification_loop.Params(
            control=None,  # noqa
            overlap=None,  # noqa
            aggregation_algorithm=None,
            task_duration_function=None,  # noqa
        ),
        markup_params=feedback_loop.Params(
            control=control.Control(
                rules=control.RuleBuilder().add_speed_control(ratio_rand=0.1, ratio_poor=0.3).rules
            ),
            assignment_check_sample=None,
            overlap=None,  # noqa
            task_duration_function=duration.get_const_task_duration_function(datetime.timedelta(seconds=20)),
        ),
        client=stub,  # noqa
        lang='RU',
    )
    assert ([markup_assignment_solutions[i] for i in [0, 1]], [markup_assignment_solutions[-1]]) == fb_loop.get_markups(
        markup_pool_id='fake', check_pool_id=None  # noqa
    )
    assert [
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    user_id='b',
                    private_comment='Fast submits, 0.1 < time <= 0.3',
                    will_expire=datetime.datetime(2020, 11, 5, 8, 0, 10),
                ),
            ),
        ),
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    user_id='c',
                    private_comment='Fast submits, <= 0.1',
                    will_expire=datetime.datetime(2020, 11, 6, 0, 0, 3),
                ),
            ),
        ),
        (
            'patch_assignment',
            (
                'm3',
                toloka.AssignmentPatch(
                    public_comment='Мало правильных решений', status=toloka.Assignment.Status.REJECTED
                ),
            ),
        ),
    ] == stub.calls



import abc
from dataclasses import dataclass
from decimal import Decimal
import json
from typing import Generic, Type, TypeVar

from pure_protobuf.dataclasses_ import field, message
import toloka.client as toloka

from ...utils import DecimalEncoder
from .common import ProtobufSerializer

TolokaObj = toloka.primitives.base.BaseTolokaObject
TolokaObjT = TypeVar('TolokaObjT', bound=Type[TolokaObj], covariant=True)


@message
@dataclass
class TolokaObject(ProtobufSerializer[TolokaObj], Generic[TolokaObjT]):
    json: str = field(1)

    @classmethod
    def serialize(cls, obj: TolokaObj):
        return cls(json=json.dumps(obj.unstructure(), cls=DecimalEncoder))

    def deserialize(self) -> TolokaObj:
        return self.toloka_cls().structure(json.loads(self.json, parse_float=Decimal))

    @staticmethod
    @abc.abstractmethod
    def toloka_cls() -> Type[TolokaObj]:
        ...


@message
class TolokaPool(TolokaObject[toloka.Pool]):
    @staticmethod
    def toloka_cls() -> Type[toloka.Pool]:
        return toloka.Pool


@message
class TolokaFilterCondition(TolokaObject[toloka.filter.FilterCondition]):
    @staticmethod
    def toloka_cls() -> Type[toloka.filter.FilterCondition]:
        return toloka.filter.FilterCondition


@message
class TolokaSkill(TolokaObject[toloka.Skill]):
    @staticmethod
    def toloka_cls() -> Type[toloka.Skill]:
        return toloka.Skill


@message
class TolokaProject(TolokaObject[toloka.Project]):
    @staticmethod
    def toloka_cls() -> Type[toloka.Project]:
        return toloka.Project


@message
class TolokaAssignment(TolokaObject[toloka.Assignment]):
    @staticmethod
    def toloka_cls() -> Type[toloka.Assignment]:
        return toloka.Assignment



from dataclasses import dataclass
from typing import List

from pure_protobuf.dataclasses_ import field, message

from ... import mapping
from .base import OptionalObject
from .common import ProtobufSerializer


@message
@dataclass
class Objects(ProtobufSerializer[mapping.Objects]):
    objects: List[OptionalObject] = field(1, default_factory=list)

    @staticmethod
    def serialize(obj: mapping.Objects) -> 'Objects':
        return Objects(objects=[OptionalObject.serialize(o) for o in obj])

    def deserialize(self) -> mapping.Objects:
        return tuple([o.deserialize() for o in self.objects])


@message
@dataclass
class TaskSingleSolution(ProtobufSerializer[mapping.TaskSingleSolution]):
    task: Objects = field(1)
    solution: Objects = field(2)

    @staticmethod
    def serialize(obj: mapping.TaskSingleSolution) -> 'TaskSingleSolution':
        task, solution = obj
        return TaskSingleSolution(task=Objects.serialize(task), solution=Objects.serialize(solution))

    def deserialize(self) -> mapping.TaskSingleSolution:
        return self.task.deserialize(), self.solution.deserialize()


@message
@dataclass
class ObjectsList(ProtobufSerializer[List[mapping.Objects]]):
    list: List[Objects] = field(1, default_factory=list)

    @staticmethod
    def serialize(obj: List[mapping.Objects]) -> 'ObjectsList':
        return ObjectsList(list=[Objects.serialize(objects) for objects in obj])

    def deserialize(self) -> List[mapping.Objects]:
        return [objects.deserialize() for objects in self.list]


@message
@dataclass
class TaskSingleSolutionList(ProtobufSerializer[List[mapping.TaskSingleSolution]]):
    list: List[TaskSingleSolution] = field(1, default_factory=list)

    @staticmethod
    def serialize(obj: List[mapping.TaskSingleSolution]) -> 'TaskSingleSolutionList':
        return TaskSingleSolutionList(list=[TaskSingleSolution.serialize(ts) for ts in obj])

    def deserialize(self) -> List[mapping.TaskSingleSolution]:
        return [ts.deserialize() for ts in self.list]



from copy import deepcopy
import datetime
import decimal
from typing import Tuple

from mock import patch
from pytest import approx
import toloka.client as toloka

from crowdom import base, classification, control, duration, evaluation, mapping, worker
from crowdom.objects import Audio, Image, Text
from . import lib


def test_collect_evaluations_from_check_assignments():
    false, true = base.BinaryEvaluation(ok=False), base.BinaryEvaluation(ok=True)
    alice, bob, peter = [worker.Human(assignment=toloka.Assignment(user_id=name)) for name in ['Alice', 'Bob', 'Peter']]
    audios = [Audio(url=f'https://storage.net/{i + 1}.wav') for i in range(3)]

    pool_input_objects = [
        (audios[0], Text(text='привет')),
        (audios[1], Text(text='до свидания')),
        (audios[2], Text(text='алло')),
    ]

    task_ids = [mapping.TaskID(input_objects) for input_objects in pool_input_objects]

    assignments_1 = [
        lib.create_check_assignment(
            pool_input_objects, [(0, True), (2, False), (None, False), (1, False)], worker_id='Alice'
        )
    ]
    assignments_2 = [
        lib.create_check_assignment(
            pool_input_objects, [(0, False), (2, True), (None, False), (1, True)], worker_id='Alice'
        ),
        lib.create_check_assignment(
            pool_input_objects, [(None, True), (2, True), (1, False), (0, False)], worker_id='Bob'
        ),
        lib.create_check_assignment(
            pool_input_objects, [(1, True), (2, True), (0, True), (None, True)], worker_id='Peter'
        ),
    ]

    for assignments, confidence_threshold, pool_input_objects, aggregation_algorithm, expected_result in (
        (
            assignments_1,
            0.6,
            pool_input_objects,
            classification.AggregationAlgorithm.MAJORITY_VOTE,
            {
                task_ids[0]: evaluation.SolutionEvaluation(ok=True, confidence=1.0, worker_labels=[(true, alice)]),
                task_ids[1]: evaluation.SolutionEvaluation(ok=False, confidence=0.0, worker_labels=[(false, alice)]),
                task_ids[2]: evaluation.SolutionEvaluation(ok=False, confidence=0.0, worker_labels=[(false, alice)]),
            },
        ),
        (
            assignments_2,
            0.6,
            pool_input_objects,
            classification.AggregationAlgorithm.MAJORITY_VOTE,
            {
                task_ids[0]: evaluation.SolutionEvaluation(
                    ok=False, confidence=1 / 3, worker_labels=[(false, alice), (false, bob), (true, peter)]
                ),
                task_ids[1]: evaluation.SolutionEvaluation(
                    ok=True, confidence=2 / 3, worker_labels=[(true, alice), (false, bob), (true, peter)]
                ),
                task_ids[2]: evaluation.SolutionEvaluation(
                    ok=True, confidence=3 / 3, worker_labels=[(true, alice), (true, bob), (true, peter)]
                ),
            },
        ),
        (
            assignments_2,
            0.7,
            pool_input_objects,
            classification.AggregationAlgorithm.MAJORITY_VOTE,
            {
                task_ids[0]: evaluation.SolutionEvaluation(
                    ok=False, confidence=1 / 3, worker_labels=[(false, alice), (false, bob), (true, peter)]
                ),
                task_ids[1]: evaluation.SolutionEvaluation(
                    ok=False, confidence=2 / 3, worker_labels=[(true, alice), (false, bob), (true, peter)]
                ),
                task_ids[2]: evaluation.SolutionEvaluation(
                    ok=True, confidence=3 / 3, worker_labels=[(true, alice), (true, bob), (true, peter)]
                ),
            },
        ),
        (
            assignments_2,
            0.7,
            # feedback loop case, no strict order for pool input objects, which are obtained from check
            # assignments (and order is obtained accordingly), in our case 3rd and 2nd objects swapped
            None,
            classification.AggregationAlgorithm.MAJORITY_VOTE,
            {
                task_ids[0]: evaluation.SolutionEvaluation(
                    ok=False, confidence=1 / 3, worker_labels=[(false, alice), (false, bob), (true, peter)]
                ),
                task_ids[2]: evaluation.SolutionEvaluation(
                    ok=True, confidence=3 / 3, worker_labels=[(true, alice), (true, bob), (true, peter)]
                ),
                task_ids[1]: evaluation.SolutionEvaluation(
                    ok=False, confidence=2 / 3, worker_labels=[(true, alice), (false, bob), (true, peter)]
                ),
            },
        ),
        (
            assignments_1,
            0.6,
            pool_input_objects,
            classification.AggregationAlgorithm.MAX_LIKELIHOOD,
            {
                task_ids[0]: evaluation.SolutionEvaluation(
                    ok=True, confidence=approx(0.8), worker_labels=[(true, alice)]
                ),
                task_ids[1]: evaluation.SolutionEvaluation(
                    ok=False, confidence=approx(0.2), worker_labels=[(false, alice)]
                ),
                task_ids[2]: evaluation.SolutionEvaluation(
                    ok=False, confidence=approx(0.2), worker_labels=[(false, alice)]
                ),
            },
        ),
        (
            assignments_2,
            0.6,
            pool_input_objects,
            classification.AggregationAlgorithm.MAX_LIKELIHOOD,
            {
                task_ids[0]: evaluation.SolutionEvaluation(
                    ok=False, confidence=approx(3 / 7), worker_labels=[(false, alice), (false, bob), (true, peter)]
                ),
                task_ids[1]: evaluation.SolutionEvaluation(
                    ok=True, confidence=approx(300 / 325), worker_labels=[(true, alice), (false, bob), (true, peter)]
                ),
                task_ids[2]: evaluation.SolutionEvaluation(
                    ok=False, confidence=approx(4 / 7), worker_labels=[(true, alice), (true, bob), (true, peter)]
                ),
            },
        ),
    ):
        assert (
            evaluation.collect_evaluations_from_check_assignments(
                assignments=assignments,
                check_task_mapping=lib.audio_transcript_check_mapping,
                pool_input_objects=pool_input_objects,
                aggregation_algorithm=aggregation_algorithm,
                confidence_threshold=confidence_threshold,
                worker_weights={'Alice': 0.8, 'Bob': 0.25, 'Peter': 0.5},
            )
            == expected_result
        )


@patch('crowdom.evaluation.shuffle')
def test_find_markup_solutions_to_check(shuffle):  # noqa
    assignments = [
        lib.create_markup_assignment(
            [
                (Audio(url='https://storage.net/01.wav'), Text(text='чвтыре')),
                (Audio(url='https://storage.net/02.wav'), Text(text='добрый день')),
                (Audio(url='https://storage.net/03.wav'), Text(text='алоэ')),
                (Audio(url='https://storage.net/04.wav'), Text(text='')),
            ]
        ),
        lib.create_markup_assignment(
            [
                # trick, same solution but other task
                (Audio(url='https://storage.net/05.wav'), Text(text='чвтыре')),
                (Audio(url='https://storage.net/06.wav'), Text(text='нет не надо')),
                (Audio(url='https://storage.net/07.wav'), Text(text='')),
                (Audio(url='https://storage.net/08.wav'), Text(text='до свидание')),
            ]
        ),
    ]
    checked_solutions = {
        mapping.TaskID((Audio(url='https://storage.net/01.wav'), Text(text='чвтыре'))),
        mapping.TaskID((Audio(url='https://storage.net/03.wav'), Text(text='алоэ'))),
        mapping.TaskID((Audio(url='https://storage.net/04.wav'), Text(text=''))),
        mapping.TaskID((Audio(url='https://storage.net/08.wav'), Text(text='до свидание'))),
        mapping.TaskID((Audio(url='https://storage.net/09.wav'), Text(text='где мой груз'))),
    }
    assert evaluation.find_markup_solutions_to_check(
        assignments, checked_solutions, lib.audio_transcript_check_mapping, check_sample=None
    ) == (
        [
            # assignment 1
            ((Audio(url='https://storage.net/02.wav'), Text(text='добрый день')), worker.Human(assignments[0][0])),
            # assignment 2
            ((Audio(url='https://storage.net/05.wav'), Text(text='чвтыре')), worker.Human(assignments[1][0])),
            ((Audio(url='https://storage.net/06.wav'), Text(text='нет не надо')), worker.Human(assignments[1][0])),
            ((Audio(url='https://storage.net/07.wav'), Text(text='')), worker.Human(assignments[1][0])),
        ],
        {
            mapping.TaskID((Audio(url='https://storage.net/01.wav'), Text(text='чвтыре'))),
            mapping.TaskID((Audio(url='https://storage.net/03.wav'), Text(text='алоэ'))),
            mapping.TaskID((Audio(url='https://storage.net/04.wav'), Text(text=''))),
            mapping.TaskID((Audio(url='https://storage.net/08.wav'), Text(text='до свидание'))),
        },
    )

    assert evaluation.find_markup_solutions_to_check(
        assignments,
        checked_solutions,
        lib.audio_transcript_check_mapping,
        check_sample=evaluation.AssignmentCheckSample(
            max_tasks_to_check=2, assignment_accuracy_finalization_threshold=None
        ),
    ) == (
        [
            # assignment 1
            ((Audio(url='https://storage.net/02.wav'), Text(text='добрый день')), worker.Human(assignments[0][0])),
            # assignment 2
            ((Audio(url='https://storage.net/05.wav'), Text(text='чвтыре')), worker.Human(assignments[1][0])),
            ((Audio(url='https://storage.net/06.wav'), Text(text='нет не надо')), worker.Human(assignments[1][0])),
        ],
        {
            mapping.TaskID((Audio(url='https://storage.net/01.wav'), Text(text='чвтыре'))),
            mapping.TaskID((Audio(url='https://storage.net/03.wav'), Text(text='алоэ'))),
            mapping.TaskID((Audio(url='https://storage.net/04.wav'), Text(text=''))),
            mapping.TaskID((Audio(url='https://storage.net/08.wav'), Text(text='до свидание'))),
        },
    )


def test_get_objects_markup_attempts():
    for with_model, expected_task_id_to_attempts in (
        (
            True,
            {
                mapping.TaskID((Audio(url='https://storage.net/01.wav'),)): 1,
                mapping.TaskID((Audio(url='https://storage.net/02.wav'),)): 2,
                mapping.TaskID((Audio(url='https://storage.net/03.wav'),)): 2,
                mapping.TaskID((Audio(url='https://storage.net/04.wav'),)): 1,
            },
        ),
        (
            False,
            {
                mapping.TaskID((Audio(url='https://storage.net/01.wav'),)): 1,
                mapping.TaskID((Audio(url='https://storage.net/02.wav'),)): 1,
                mapping.TaskID((Audio(url='https://storage.net/03.wav'),)): 1,
                mapping.TaskID((Audio(url='https://storage.net/04.wav'),)): 0,
            },
        ),
    ):
        assert (
            evaluation.get_tasks_attempts(
                [
                    lib.create_markup_assignment(
                        [
                            (Audio(url='https://storage.net/01.wav'), Text(text='привет')),
                            (Audio(url='https://storage.net/02.wav'), Text(text='')),
                            (Audio(url='https://storage.net/03.wav'), Text(text='алло')),
                        ]
                    ),
                    lib.create_markup_assignment(
                        [
                            (Audio(url='https://storage.net/02.wav'), Text(text='?')),
                            (Audio(url='https://storage.net/03.wav'), Text(text='але')),
                            (Audio(url='https://storage.net/04.wav'), Text(text='нет не надо')),
                        ],
                        id='',  # model worker
                    ),
                ],
                lib.audio_transcript_mapping,
                with_model,
            )
            == expected_task_id_to_attempts
        )


def test_evaluate_markup_assignment():
    assignment, solutions = lib.create_markup_assignment(
        audio_text_pairs=[
            (Audio(url='https://storage.net/01.wav'), Text(text='алло')),
            (Audio(url='https://storage.net/02.wav'), Text(text='михаил')),
            (Audio(url='https://storage.net/03.wav'), Text(text='до свидания')),
            (Audio(url='https://storage.net/04.wav'), Text(text='нет не нужно')),
            (Audio(url='https://storage.net/05.wav'), Text(text='')),
        ],
        id='123',
        user_id='gleb',
    )
    assert evaluation.CheckAssignmentAccuracyEvaluationStrategy(
        solution_id_to_evaluation=lib.generate_evaluations(
            [
                # audio matches, but solution differs
                ('https://storage.net/01.wav', '? алло', True),
                # match, BAD
                ('https://storage.net/01.wav', 'алло', False),
                # audio matches, but solution differs
                ('https://storage.net/02.wav', '?', True),
                # match, OK
                ('https://storage.net/03.wav', 'до свидания', True),
                # audio not matches
                ('https://storage.net/00.wav', 'как', False),
                # no checks for 04.wav
                # match, BAD
                ('https://storage.net/05.wav', '', False),
            ]
        ),
        check_task_mapping=lib.audio_transcript_check_mapping,
    ).evaluate_assignment((assignment, solutions)) == evaluation.AssignmentEvaluation(
        assignment=assignment, ok_checks=1, total_checks=3, objects=5, incorrect_solution_indexes=[0, 4]
    )


def test_evaluate_classification_assignment():
    # TODO: need to test optional solution objects
    dog, cat = lib.dog, lib.cat
    images = [Image(url=f'https://storage.net/{i + 1}.jpg') for i in range(3)]
    control_images = [Image(url=f'https://storage.net/c_{i + 1}.jpg') for i in range(3)]
    assignment, solutions = lib.create_classification_assignment(
        image_class_solutions=[
            (control_images[1], dog),  # BAD (cat)
            (images[0], cat),
            (images[1], dog),
            (control_images[2], dog),  # OK
            (control_images[0], cat),  # BAD (dog)
            (images[2], cat),
        ],
        image_class_controls=[(control_images[0], dog), (control_images[1], cat), (control_images[2], dog)],
        user_id='bob',
        id='321',
    )
    assert evaluation.ControlTasksAssignmentAccuracyEvaluationStrategy(
        task_mapping=lib.image_classification_mapping,
    ).evaluate_assignment((assignment, solutions)) == evaluation.AssignmentEvaluation(
        assignment=assignment, ok_checks=1, total_checks=3, objects=6, incorrect_solution_indexes=[0, 4]
    )


@patch('crowdom.control.rule.datetime', wraps=datetime)
def test_apply_rules_to_assignment(mock_datetime):
    now = datetime.datetime(2020, 11, 5)
    mock_datetime.datetime.now.return_value = now
    total_checks = 3
    control_params = control.Control(
        control.RuleBuilder().add_static_reward(threshold=0.5).add_control_task_control(total_checks, 2, 3).build()
    )
    set_verdict_rules = control_params.filter_rules(control.AssignmentAccuracyPredicate, control.SetAssignmentStatus)
    block_rules = control_params.filter_rules(control.AssignmentAccuracyPredicate, control.BlockUser)
    for incorrect_solution_indexes, expected_toloka_calls in (  # add case
        (
            [0, 1],
            [
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='gleb',
                            pool_id='fake',
                            private_comment='Control tasks: [0, 2) done correctly',
                            will_expire=datetime.datetime(2020, 11, 5) + datetime.timedelta(hours=8),
                        ),
                    ),
                ),
                (
                    'patch_assignment',
                    (
                        '123',
                        toloka.assignment.AssignmentPatch(
                            status=toloka.Assignment.REJECTED,
                            public_comment='Тапсырмаларды нөмірлермен тексеріңіз: 1, 2. Тапсырмаларды тексеру және '
                            'апелляцияларды беру туралы толығырақ жоба нұсқаулығынан оқыңыз.',
                        ),
                    ),
                ),
            ],
        ),
        (
            [0],
            [
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='gleb',
                            pool_id='fake',
                            private_comment='Control tasks: [2, 3) done correctly',
                            will_expire=datetime.datetime(2020, 11, 5) + datetime.timedelta(hours=1),
                        ),
                    ),
                ),
                (
                    'patch_assignment',
                    ('123', toloka.assignment.AssignmentPatch(status=toloka.Assignment.ACCEPTED, public_comment='')),
                ),
            ],
        ),
        (
            [],
            [
                (
                    'patch_assignment',
                    ('123', toloka.assignment.AssignmentPatch(status=toloka.Assignment.ACCEPTED, public_comment='')),
                )
            ],
        ),
    ):
        ok_checks = total_checks - len(incorrect_solution_indexes)
        stub = lib.TolokaClientCallRecorderStub()
        evaluation.apply_rules_to_assignment(
            set_verdict_rules=set_verdict_rules,
            block_rules=block_rules,
            evaluation=evaluation.AssignmentEvaluation(
                assignment=toloka.Assignment(id='123', user_id='gleb'),
                total_checks=total_checks,
                ok_checks=ok_checks,
                objects=3,
                incorrect_solution_indexes=incorrect_solution_indexes,
            ),
            client=stub,  # noqa
            lang='KK',
            pool_id='fake',
        )
        assert stub.calls == expected_toloka_calls


@patch('crowdom.control.rule.datetime', wraps=datetime)
def test_evaluate_submitted_markup_assignments_and_apply_rules(mock_datetime):
    now = datetime.datetime(2020, 11, 5)
    mock_datetime.datetime.now.return_value = now
    assignment_1 = lib.create_markup_assignment(
        audio_text_pairs=[
            # unknown
            (Audio(url='https://storage.net/00.wav'), Text(text='здравствуйте')),
            # evaluations_1 OK
            (Audio(url='https://storage.net/01.wav'), Text(text='нет не надо')),
            # evaluations_1 BAD
            (Audio(url='https://storage.net/02.wav'), Text(text='до свидания')),
            # unknown
            (Audio(url='https://storage.net/03.wav'), Text(text='алло а да')),
            # evaluations_1 OK
            (Audio(url='https://storage.net/04.wav'), Text(text='улица гончарова')),
            # unknown
            (Audio(url='https://storage.net/05.wav'), Text(text='нет')),
            # evaluations_1 BAD
            (Audio(url='https://storage.net/06.wav'), Text(text='')),
            # unknown
            (Audio(url='https://storage.net/07.wav'), Text(text='да')),
        ],
        id='assignment 0',
        user_id='user 0',
    )
    assignment_2 = lib.create_markup_assignment(
        audio_text_pairs=[
            # unknown
            (Audio(url='https://storage.net/00.wav'), Text(text='здравствуйте')),
            # evaluations_1 BAD
            (Audio(url='https://storage.net/01.wav'), Text(text='нет')),
            # unknown
            (Audio(url='https://storage.net/02.wav'), Text(text='')),
            # evaluations_1 OK
            (Audio(url='https://storage.net/03.wav'), Text(text='а да')),
            # evaluations_1 BAD
            (Audio(url='https://storage.net/04.wav'), Text(text='улица горчакова')),
        ],
        id='assignment 1',
        user_id='user 1',
    )
    assignment_model = deepcopy(assignment_2)
    assignment_model[0].id = ''
    assignment_model[0].user_id = 'model'

    evaluations_1 = lib.generate_evaluations(
        [
            ('https://storage.net/01.wav', 'нет', False),
            ('https://storage.net/01.wav', 'нет не надо', True),
            ('https://storage.net/02.wav', 'до свидания', False),
            ('https://storage.net/03.wav', 'а', False),
            ('https://storage.net/03.wav', 'а да', True),
            ('https://storage.net/04.wav', 'улица гончарова', True),
            ('https://storage.net/04.wav', 'улица горчакова', False),
            ('https://storage.net/06.wav', '', False),
        ]
    )

    for (
        assignments,
        min_correct_solutions_ratio_for_acceptance,
        solution_id_to_evaluation,
        expected_toloka_calls,
        expected_statuses,
    ) in (
        (
            [assignment_1, assignment_2],
            0.5,
            evaluations_1,
            [
                (
                    'patch_assignment',
                    (
                        'assignment 0',
                        toloka.assignment.AssignmentPatch(
                            status=toloka.Assignment.ACCEPTED,
                            public_comment='',
                        ),
                    ),
                ),
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='user 1',
                            pool_id='fake',
                            private_comment='Control tasks: [1, 2) done correctly',
                            will_expire=datetime.datetime(2020, 11, 5) + datetime.timedelta(hours=1),
                        ),
                    ),
                ),
                (
                    'patch_assignment',
                    (
                        'assignment 1',
                        toloka.assignment.AssignmentPatch(
                            status=toloka.Assignment.REJECTED,
                            public_comment='Проверьте задания c номерами: 2, 5. Подробнее о проверке заданий и подаче апелляций'
                            ' читайте в инструкции проекта.',
                        ),
                    ),
                ),
            ],
            [toloka.Assignment.ACCEPTED, toloka.Assignment.REJECTED],
        ),
        (
            [assignment_1],
            0.6,
            evaluations_1,
            [
                (
                    'patch_assignment',
                    (
                        'assignment 0',
                        toloka.assignment.AssignmentPatch(
                            status=toloka.Assignment.REJECTED,
                            public_comment='Проверьте задания c номерами: 3, 7. Подробнее о проверке заданий и подаче апелляций'
                            ' читайте в инструкции проекта.',
                        ),
                    ),
                ),
            ],
            [toloka.Assignment.REJECTED],
        ),
        (
            [assignment_2],
            0.3,
            evaluations_1,
            [
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='user 1',
                            pool_id='fake',
                            private_comment='Control tasks: [1, 2) done correctly',
                            will_expire=datetime.datetime(2020, 11, 5) + datetime.timedelta(hours=1),
                        ),
                    ),
                ),
                (
                    'patch_assignment',
                    (
                        'assignment 1',
                        toloka.assignment.AssignmentPatch(
                            status=toloka.Assignment.ACCEPTED,
                            public_comment='',
                        ),
                    ),
                ),
            ],
            [toloka.Assignment.ACCEPTED],
        ),
        (
            [assignment_model],
            0.3,
            evaluations_1,
            [],
            [toloka.Assignment.ACCEPTED],
        ),
    ):
        stub = lib.TolokaClientCallRecorderStub()
        actual_statuses = evaluation.evaluate_submitted_assignments_and_apply_rules(
            submitted_assignments=assignments,
            assignment_accuracy_evaluation_strategy=evaluation.CheckAssignmentAccuracyEvaluationStrategy(
                solution_id_to_evaluation=solution_id_to_evaluation,
                check_task_mapping=lib.audio_transcript_check_mapping,
            ),
            control_params=control.Control(
                rules=control.RuleBuilder()
                .add_static_reward(threshold=min_correct_solutions_ratio_for_acceptance)
                # control task count is not exact for all test cases
                .add_control_task_control(4, 1, 2)
                .build()
            ),
            client=stub,  # noqa
            lang='RU',
            pool_id='fake',
        )
        assert actual_statuses == expected_statuses
        assert stub.calls == expected_toloka_calls
        for assignment, expected_status in zip(assignments, expected_statuses):
            if not assignment[0].id:
                # statuses for model assignments should be set on the fly
                assert assignment[0].status == expected_status


def test_get_markup_task_id():
    check_task = mapping.TaskID((Audio(url='https://storage.net/1.wav'), Text(text='привет')))
    assert check_task.id == "Audio(url='https://storage.net/1.wav') Text(text='привет')"
    assert (
        evaluation.get_markup_task_id(check_task, lib.audio_transcript_mapping).id
        == "Audio(url='https://storage.net/1.wav')"
    )


def test_find_finalized_objects():
    assignments = [
        # Finalize by accuracy: NO
        lib.create_markup_assignment(
            audio_text_pairs=[
                # attempt 1, UNKNOWN
                (Audio(url='https://storage.net/00.wav'), Text(text='здравствуйте')),
                # attempt 1, BAD
                (Audio(url='https://storage.net/01.wav'), Text(text='не надо')),
                # attempt 1, OK (finalized by quality)
                (Audio(url='https://storage.net/02.wav'), Text(text='до свидания')),
                # attempt 1, BAD
                (Audio(url='https://storage.net/03.wav'), Text(text='? а да')),
                # attempt 1, BAD
                (Audio(url='https://storage.net/04.wav'), Text(text='улица горчакова')),
                # attempt 1, UNKNOWN
                (Audio(url='https://storage.net/05.wav'), Text(text='нет')),
            ],
            id='assignment 0',
            user_id='user 0',
        ),
        # Finalize by accuracy: YES
        lib.create_markup_assignment(
            audio_text_pairs=[
                # attempt 1, OK (finalized by quality)
                (Audio(url='https://storage.net/06.wav'), Text(text='')),
                # attempt 1, UNKNOWN (finalized by assignment accuracy)
                (Audio(url='https://storage.net/07.wav'), Text(text='где ?')),
                # attempt 1, BAD
                (Audio(url='https://storage.net/08.wav'), Text(text='хорошо довайте')),
                # attempt 1, BAD
                (Audio(url='https://storage.net/09.wav'), Text(text='переключи на оратора')),
                # attempt 1, OK (finalized by quality)
                (Audio(url='https://storage.net/10.wav'), Text(text='леруа мерлен')),
                # attempt 1, UNKNOWN (finalized by assignment accuracy)
                (Audio(url='https://storage.net/11.wav'), Text(text='угу')),
            ],
            id='assignment 1',
            user_id='user 0',
        ),
        # Finalize by accuracy: NO
        lib.create_markup_assignment(
            audio_text_pairs=[
                # attempt 2, UNKNOWN
                (Audio(url='https://storage.net/03.wav'), Text(text='кхм да')),
                # attempt 2, OK (finalized by quality)
                (Audio(url='https://storage.net/05.wav'), Text(text='что нет')),
                # attempt 2, BAD
                (Audio(url='https://storage.net/08.wav'), Text(text='? давайте')),
                # attempt 1, BAD
                (Audio(url='https://storage.net/12.wav'), Text(text='два чеыре семь')),
                # attempt 1, UNKNOWN
                (Audio(url='https://storage.net/13.wav'), Text(text='приезжайте потом')),
                # attempt 1, BAD
                (Audio(url='https://storage.net/14.wav'), Text(text='позовите ? алену')),
            ],
            id='assignment 2',
            user_id='user 1',
        ),
        # Finalize by accuracy: NO
        lib.create_markup_assignment(
            audio_text_pairs=[
                # attempt 3, BAD (finalized by max attempts)
                (Audio(url='https://storage.net/03.wav'), Text(text='? а да')),
                # attempt 3, UNKNOWN (finalized by max attempts)
                (Audio(url='https://storage.net/08.wav'), Text(text='да давайте')),
                # attempt 2, BAD
                (Audio(url='https://storage.net/13.wav'), Text(text='прижайте потм')),
                # attempt 2, OK (finalized by quality)
                (Audio(url='https://storage.net/14.wav'), Text(text='позовите алёну')),
                # attempt 1, BAD
                (Audio(url='https://storage.net/15.wav'), Text(text='')),
                # attempt 1, UNKNOWN
                (Audio(url='https://storage.net/16.wav'), Text(text='')),
            ],
            id='assignment 3',
            user_id='user 2',
        ),
    ]

    evaluations = lib.generate_evaluations(
        [
            # assignment 0
            ('https://storage.net/01.wav', 'не надо', False),
            ('https://storage.net/02.wav', 'до свидания', True),
            ('https://storage.net/03.wav', '? а да', False),
            ('https://storage.net/04.wav', 'улица горчакова', False),
            # assignment 1
            ('https://storage.net/06.wav', '', True),
            ('https://storage.net/08.wav', 'хорошо довайте', False),
            ('https://storage.net/09.wav', 'переключи на оратора', False),
            ('https://storage.net/10.wav', 'леруа мерлен', True),
            # assignment 2
            ('https://storage.net/05.wav', 'что нет', True),
            ('https://storage.net/08.wav', '? давайте', False),
            ('https://storage.net/12.wav', 'два чеыре семь', False),
            ('https://storage.net/14.wav', 'позовите ? алену', False),
            # assignment 3
            # already present: ('https://storage.net/03.wav', '? а да', False),
            ('https://storage.net/13.wav', 'прижайте потм', False),
            ('https://storage.net/14.wav', 'позовите алёну', True),
            ('https://storage.net/15.wav', '', False),
        ]
    )

    finalized_task_ids = evaluation.find_finalized_tasks(
        markup_assignments=assignments,
        solution_id_to_evaluation=evaluations,
        markup_task_mapping=lib.audio_transcript_mapping,
        check_task_mapping=lib.audio_transcript_check_mapping,
        check_sample=evaluation.AssignmentCheckSample(
            assignment_accuracy_finalization_threshold=0.5, max_tasks_to_check=4
        ),
        max_object_markup_attempts=3,
    )

    assert {task.id for task in finalized_task_ids} == set(
        f"Audio(url='https://storage.net/{i:02d}.wav')" for i in [2, 6, 7, 10, 11, 5, 3, 8, 14]
    )


def test_ok_objects_ratio():
    assert (
        evaluation.get_ok_tasks_ratio(
            lib.generate_evaluations(
                [
                    # OK
                    ('https://storage.net/01.wav', 'понятно до свидания', True),
                    ('https://storage.net/01.wav', 'понятно да увидимся', False),
                    # BAD
                    ('https://storage.net/02.wav', 'до свидания', False),
                    # OK
                    ('https://storage.net/03.wav', 'а', True),
                    # BAD
                    ('https://storage.net/04.wav', 'улица гончарова', False),
                    ('https://storage.net/04.wav', 'улица горчакова', False),
                    # OK
                    ('https://storage.net/05.wav', '', True),
                ]
            ),
            lib.audio_transcript_mapping,
        )
        == 3 / 5
    )


def test_calculate_bonuses_for_markup_assignment():
    rules = [
        control.Rule(
            predicate=control.PredicateExpression(
                boolean_operator=control.BooleanOperator.AND,
                predicates=[
                    control.AssignmentAccuracyPredicate(threshold=0.5, comparison=control.ComparisonType('>')),
                    control.AssignmentAccuracyPredicate(threshold=0.7, comparison=control.ComparisonType('<')),
                ],
            ),
            action=control.GiveBonusToUser(amount_usd=0.03),
        ),
        control.Rule(
            predicate=control.PredicateExpression(
                boolean_operator=control.BooleanOperator.AND,
                predicates=[
                    control.AssignmentAccuracyPredicate(threshold=0.7, comparison=control.ComparisonType('>=')),
                    control.AssignmentAccuracyPredicate(threshold=0.9, comparison=control.ComparisonType('<')),
                ],
            ),
            action=control.GiveBonusToUser(amount_usd=0.04),
        ),
        control.Rule(
            predicate=control.PredicateExpression(
                boolean_operator=control.BooleanOperator.AND,
                predicates=[
                    control.AssignmentAccuracyPredicate(threshold=0.9, comparison=control.ComparisonType('>=')),
                    control.AssignmentAccuracyPredicate(threshold=1.0, comparison=control.ComparisonType('<=')),
                ],
            ),
            action=control.GiveBonusToUser(amount_usd=0.1),
        ),
    ]

    for assignment_evaluation, expected_bonuses in (
        (
            evaluation.AssignmentEvaluation(
                assignment=toloka.Assignment(id='assignment0', user_id='user1'),
                incorrect_solution_indexes=[],
                ok_checks=9,
                total_checks=10,
                objects=10,
            ),
            [
                toloka.user_bonus.UserBonus(
                    user_id='user1', assignment_id='assignment0', amount=decimal.Decimal(0.1), without_message=True
                ),
            ],
        ),
        (
            evaluation.AssignmentEvaluation(
                assignment=toloka.Assignment(id='assignment1', user_id='user2'),
                incorrect_solution_indexes=[],
                ok_checks=8,
                total_checks=10,
                objects=10,
            ),
            [
                toloka.user_bonus.UserBonus(
                    user_id='user2', assignment_id='assignment1', amount=decimal.Decimal(0.04), without_message=True
                )
            ],
        ),
        (
            evaluation.AssignmentEvaluation(
                assignment=toloka.Assignment(id='assignment2', user_id='user3'),
                incorrect_solution_indexes=[],
                ok_checks=5,
                total_checks=10,
                objects=10,
            ),
            [],
        ),
    ):
        assert evaluation.calculate_bonuses_for_markup_assignment(rules, assignment_evaluation) == expected_bonuses


def test_process_assignment_by_duration():
    task_duration_hint = datetime.timedelta(seconds=10)
    control_params = control.Control(
        rules=control.RuleBuilder().add_speed_control(ratio_rand=0.1, ratio_poor=0.6).rules
    )

    reject_rules = control_params.filter_rules(
        predicate_type=control.AssignmentDurationPredicate, action_type=control.SetAssignmentStatus
    )

    block_rules = control_params.filter_rules(
        predicate_type=control.AssignmentDurationPredicate, action_type=control.BlockUser
    )

    assignment_start = datetime.datetime(year=2020, month=11, day=3)

    image = Image(url='https://storage.net/1.jpg')
    dog = lib.dog
    tasks_count = 4

    for assignment_duration, expected_verdict, expected_toloka_calls in (
        (
            datetime.timedelta(seconds=1) * tasks_count,
            toloka.Assignment.Status.REJECTED,
            [
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='fake_user',
                            private_comment='Fast submits, <= 0.1',
                            will_expire=datetime.datetime(year=2020, month=11, day=4, second=4),
                            pool_id='fake_pool',
                        ),
                    ),
                ),
                (
                    'patch_assignment',
                    (
                        'fake',
                        toloka.AssignmentPatch(
                            public_comment='Мало правильных решений', status=toloka.Assignment.Status.REJECTED
                        ),
                    ),
                ),
            ],
        ),
        (
            datetime.timedelta(seconds=6) * tasks_count,
            None,
            [
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='fake_user',
                            private_comment='Fast submits, 0.1 < time <= 0.6',
                            will_expire=datetime.datetime(year=2020, month=11, day=3, hour=8, second=24),
                            pool_id='fake_pool',
                        ),
                    ),
                )
            ],
        ),
        (datetime.timedelta(seconds=7) * tasks_count, None, []),
    ):
        assignment = lib.create_classification_assignment(
            image_class_solutions=[(image, dog) for _ in range(tasks_count)],
            image_class_controls=[],
            duration=assignment_duration,
            assignment_start=assignment_start,
            user_id='fake_user',
            pool_id='fake_pool',
        )[0]
        stub = lib.TolokaClientCallRecorderStub()
        verdict = evaluation.process_assignment_by_duration(
            reject_rules=reject_rules,
            block_rules=block_rules,
            assignment=assignment,
            lang='RU',
            client=stub,
            assignment_duration_hint=task_duration_hint * len(assignment.tasks),
        )  # noqa
        assert stub.calls == expected_toloka_calls
        assert verdict == expected_verdict


def test_prior_filter_assignments():
    task_duration_hint = datetime.timedelta(seconds=10)
    control_params = control.Control(
        rules=control.RuleBuilder().add_speed_control(ratio_rand=0.1, ratio_poor=0.6).rules
    )

    assignment_start = datetime.datetime(year=2020, month=11, day=3)

    images = [Image(url=str(i)) for i in range(5)]
    dog = lib.dog

    assignments = [
        lib.create_classification_assignment(
            image_class_solutions=[(images[i], dog) for i in range(tasks_count)],
            image_class_controls=[],
            duration=datetime.timedelta(seconds=assignment_duration_seconds),
            assignment_start=assignment_start,
            user_id=f'duration-{assignment_duration_seconds}-tasks-{tasks_count}',
            id=f'duration-{assignment_duration_seconds}-tasks-{tasks_count}',
            pool_id='fake_pool',
        )
        for assignment_duration_seconds, tasks_count in zip([4, 5, 24, 25, 7, 7], [4] * 4 + [2, 1])
    ]

    def custom_task_duration_function(input_objects: Tuple[Image]) -> datetime.timedelta:
        image = input_objects[0]
        return task_duration_hint * (int(image.url) + 1)

    for task_duration_function, expected_assignments, expected_fast_assignments, expected_calls in [
        (
            duration.get_const_task_duration_function(task_duration_hint),
            [assignments[i] for i in range(1, 6)],
            [assignments[0]],
            [
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='duration-4-tasks-4',
                            private_comment='Fast submits, <= 0.1',
                            will_expire=datetime.datetime(year=2020, month=11, day=4, second=4),
                            pool_id='fake_pool',
                        ),
                    ),
                ),
                (
                    'patch_assignment',
                    (
                        'duration-4-tasks-4',
                        toloka.AssignmentPatch(
                            public_comment='Мало правильных решений', status=toloka.Assignment.Status.REJECTED
                        ),
                    ),
                ),
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='duration-5-tasks-4',
                            private_comment='Fast submits, 0.1 < time <= 0.6',
                            will_expire=datetime.datetime(year=2020, month=11, day=3, hour=8, second=5),
                            pool_id='fake_pool',
                        ),
                    ),
                ),
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='duration-24-tasks-4',
                            private_comment='Fast submits, 0.1 < time <= 0.6',
                            will_expire=datetime.datetime(year=2020, month=11, day=3, hour=8, second=24),
                            pool_id='fake_pool',
                        ),
                    ),
                ),
                # incomplete assignment, no reject, but restrict
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='duration-7-tasks-2',
                            private_comment='Fast submits, 0.1 < time <= 0.6',
                            will_expire=datetime.datetime(year=2020, month=11, day=3, hour=8, second=7),
                            pool_id='fake_pool',
                        ),
                    ),
                ),
            ],
        ),
        (
            custom_task_duration_function,
            [assignments[i] for i in [2, 3, 4, 5]],
            [assignments[i] for i in [0, 1]],
            [
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='duration-4-tasks-4',
                            private_comment='Fast submits, <= 0.1',
                            will_expire=datetime.datetime(year=2020, month=11, day=4, second=4),
                            pool_id='fake_pool',
                        ),
                    ),
                ),
                (
                    'patch_assignment',
                    (
                        'duration-4-tasks-4',
                        toloka.AssignmentPatch(
                            public_comment='Мало правильных решений', status=toloka.Assignment.Status.REJECTED
                        ),
                    ),
                ),
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='duration-5-tasks-4',
                            private_comment='Fast submits, <= 0.1',
                            will_expire=datetime.datetime(year=2020, month=11, day=4, hour=0, second=5),
                            pool_id='fake_pool',
                        ),
                    ),
                ),
                (
                    'patch_assignment',
                    (
                        'duration-5-tasks-4',
                        toloka.AssignmentPatch(
                            public_comment='Мало правильных решений', status=toloka.Assignment.Status.REJECTED
                        ),
                    ),
                ),
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='duration-24-tasks-4',
                            private_comment='Fast submits, 0.1 < time <= 0.6',
                            will_expire=datetime.datetime(year=2020, month=11, day=3, hour=8, second=24),
                            pool_id='fake_pool',
                        ),
                    ),
                ),
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='duration-25-tasks-4',
                            private_comment='Fast submits, 0.1 < time <= 0.6',
                            will_expire=datetime.datetime(year=2020, month=11, day=3, hour=8, second=25),
                            pool_id='fake_pool',
                        ),
                    ),
                ),
                (
                    'set_user_restriction',
                    (
                        toloka.user_restriction.PoolUserRestriction(
                            user_id='duration-7-tasks-2',
                            private_comment='Fast submits, 0.1 < time <= 0.6',
                            will_expire=datetime.datetime(year=2020, month=11, day=3, hour=8, second=7),
                            pool_id='fake_pool',
                        ),
                    ),
                ),
            ],
        ),
    ]:
        stub = lib.TolokaClientCallRecorderStub()

        filtered_assignments, fast_assignments = evaluation.prior_filter_assignments(
            client=stub,  # noqa
            submitted_assignments=assignments,
            control_params=control_params,
            lang='RU',
            task_duration_function=task_duration_function,
        )

        assert filtered_assignments == expected_assignments
        assert fast_assignments == expected_fast_assignments
        assert stub.calls == expected_calls


def test_calculate_worker_weights():
    dog, cat = lib.dog, lib.cat
    images = [Image(url=f'https://storage.net/{i}.jpg') for i in range(6)]
    control_images = [(Image(url=f'https://storage.net/{i}_control.jpg'), cat if i % 2 == 0 else dog) for i in range(4)]
    assignments = [
        lib.create_classification_assignment(
            image_class_pairs, control_images, user_id=user_id, duration=datetime.timedelta(seconds=duration)
        )
        for image_class_pairs, user_id, duration in [
            #   john 2 bad, 2 ok
            (
                [(images[0], cat), (control_images[1][0], cat), (images[3], dog), (control_images[2][0], cat)],
                'john',
                35,
            ),
            ([(control_images[2][0], cat), (images[5], dog), (images[1], dog), (control_images[0][0], dog)], 'john', 3),
            #   bob  4 bad, 0 ok
            ([(control_images[0][0], dog), (images[2], cat), (images[1], dog), (control_images[2][0], dog)], 'bob', 31),
            ([(images[4], dog), (control_images[3][0], cat), (control_images[1][0], cat), (images[0], cat)], 'bob', 20),
            #   mary  2 bad, 0 ok
            (
                [(images[5], cat), (images[4], dog), (control_images[1][0], cat), (control_images[3][0], cat)],
                'mary',
                39,
            ),
            #   alice 0 bad, 2 ok
            (
                [(images[2], dog), (images[3], cat), (control_images[2][0], cat), (control_images[0][0], cat)],
                'alice',
                22,
            ),
        ]
    ]

    assert evaluation.calculate_worker_weights(
        assignments,
        evaluation.ControlTasksAssignmentAccuracyEvaluationStrategy(lib.image_classification_mapping),
    ) == {'alice': 2.5 / 3, 'bob': 0.5 / 5, 'john': 2.5 / 5, 'mary': 0.5 / 3}



import datetime
import itertools
from typing import List, Union, Iterable, Dict, Tuple

from pytest import approx
import toloka.client as toloka

from crowdom import (
    classification,
    classification_loop,
    control,
    duration,
    evaluation,
    mapping,
    pool as pool_config,
    worker,
)
from crowdom.objects import Audio, Image, Text
from . import lib


def test_loop_e2e():
    assignment_start = datetime.datetime(year=2020, month=10, day=5, hour=13, minute=15)

    project = toloka.Project(
        id='project',
        task_spec=toloka.project.TaskSpec(
            input_spec=lib.image_classification_mapping.to_toloka_input_spec(),
            output_spec=lib.image_classification_mapping.to_toloka_output_spec(),
        ),
    )

    class TolokaClientStub(lib.TolokaClientIntegrationStub):
        assignments_by_iterations: List[List[toloka.Assignment]]
        assignments_requests: int
        id_to_assignment: Dict[str, toloka.Assignment]

        def __init__(self, assignments_by_iterations: List[List[toloka.Assignment]]):
            self.assignments_by_iterations = assignments_by_iterations
            self.assignments_requests = 0
            super(TolokaClientStub, self).__init__(list(itertools.chain(*assignments_by_iterations)), [project])

        # Эмуляция выполнения заданий толокерами.
        # Логика этого метода сильно зависит от текущей реализации classification loop.
        def get_assignments(
            self,
            status: Union[toloka.Assignment.Status, List[toloka.Assignment.Status]],
            pool_id: str,
        ) -> List[toloka.Assignment]:
            if not isinstance(status, list):
                status = [status]
            if status == [toloka.Assignment.SUBMITTED]:
                # Каждая итерация цикла содержит 3 запроса к страницам заданий.
                loop_iteration = self.assignments_requests // 3
                assignments = self.assignments_by_iterations[loop_iteration]
            else:
                assignments = [
                    assignment for assignment in self.id_to_assignment.values() if assignment.status in status
                ]
            self.assignments_requests += 1
            return assignments

    overlap = 2

    task_duration_hint = datetime.timedelta(seconds=10)
    control_tasks_count = 2
    real_tasks_count = 2
    ratio_rand, ratio_poor = 0.1, 0.3

    control_params = control.Control(
        rules=control.RuleBuilder()
        .add_static_reward(threshold=0.5)
        .add_speed_control(ratio_rand, ratio_poor)
        .add_control_task_control(control_tasks_count, 1, 2)
        .build()
    )

    pool_id = 'fake pool'
    pool_cfg = pool_config.ClassificationConfig(
        project_id=project.id,
        private_name=pool_id,
        reward_per_assignment=0.01,
        task_duration_hint=task_duration_hint,
        real_tasks_count=real_tasks_count,
        control_tasks_count=control_tasks_count,
        overlap=overlap,
        control_params=control_params,
    )

    dog, cat, crow = lib.dog, lib.cat, lib.crow
    images = [Image(url=f'https://storage.net/{i}.jpg') for i in range(6)]
    control_images = [(Image(url=f'https://storage.net/{i}_control.jpg'), cat if i % 2 == 0 else dog) for i in range(4)]
    # 2 real tasks, 2 control tasks, 1/2 correct control tasks for accept
    # static overlap 2
    assignments = [
        # iteration 0
        [
            lib.create_classification_assignment(
                image_class_pairs,
                control_images,
                user_id=user_id,
                duration=datetime.timedelta(seconds=duration),
                assignment_start=assignment_start,
                pool_id=pool_id,
            )[0]
            for image_class_pairs, user_id, duration in [
                #   john  [(0.jpg cat) (1_control.jpg cat [bad]) (3.jpg cat) (2_control.jpg cat [ok])]  ACCEPTED
                (
                    [(images[0], cat), (control_images[1][0], cat), (images[3], cat), (control_images[2][0], cat)],
                    'john',
                    35,
                ),
                #   bob   [(0_control.jpg dog [bad]) (2.jpg cat) (1.jpg dog) (2_control.jpg dog [bad])] REJECTED
                (
                    [(control_images[0][0], dog), (images[2], cat), (images[1], dog), (control_images[2][0], dog)],
                    'bob',
                    31,
                ),
                #   mary  [(5.jpg cat) (4.jpg dog) (1_control.jpg cat [bad]) (3_control.jpg cat [bad])] REJECTED
                (
                    [(images[5], cat), (images[4], dog), (control_images[1][0], cat), (control_images[3][0], cat)],
                    'mary',
                    39,
                ),
                #   bob   [(4.jpg dog) (3_control.jpg cat [bad]) (1_control.jpg cat [bad]) (0.jpg cat)] REJECTED
                (
                    [(images[4], dog), (control_images[3][0], cat), (control_images[1][0], cat), (images[0], cat)],
                    'bob',
                    20,
                ),
                #   alice [(2.jpg dog) (3.jpg cat) (2_control.jpg cat [ok]) (0_control.jpg cat [ok])]   ACCEPTED
                (
                    [(images[2], dog), (images[3], cat), (control_images[2][0], cat), (control_images[0][0], cat)],
                    'alice',
                    22,
                ),
                #   john  [(2_control.jpg dog [ok]) (5.jpg dog) (1.jpg dog) (0_control.jpg dog [bad])]
                #   control tasks OK (1/2), but REJECTED BY PRIOR
                (
                    [(control_images[2][0], cat), (images[5], dog), (images[1], dog), (control_images[0][0], dog)],
                    'john',
                    3,
                ),
            ]
        ],
        # result (BEFORE -> SUBMITTED -> ACCEPTED -> RESULT):
        #   0.jpg () -> (cat|john, cat|bob) -> (cat|john), 0.5 conf
        #   1.jpg () -> (dog|bob, dog|john) -> ()
        #   2.jpg () -> (cat|bob, dog|alice) -> (dog|alice), 0.83 conf, but only one answer
        #   3.jpg () -> (cat|john, cat|alice) -> (cat|john, cat|alice), 0.90 conf
        #   4.jpg () -> (dog|mary, dog|bob) -> ()
        #   5.jpg () -> (cat|mary, dog|john) -> ()
        #
        #   mary, bob and john are blocked now
        #
        # iteration 1
        [
            lib.create_classification_assignment(
                image_class_pairs,
                control_images,
                user_id=user_id,
                duration=datetime.timedelta(seconds=duration),
                assignment_start=assignment_start,
                pool_id=pool_id,
            )[0]
            for image_class_pairs, user_id, duration in [
                #   martha  [(0_control.jpg dog [ok]) (5.jpg cat) (4.jpg dog) (2_control.jpg dog [ok])]
                #   control tasks OK (2/2), but REJECTED BY PRIOR
                (
                    [(control_images[0][0], cat), (images[5], cat), (images[4], dog), (control_images[2][0], cat)],
                    'martha',
                    2,
                ),
                #   fedor [(1.jpg dog) (5.jpg cat) (1_control.jpg dog [ok]) (2_control.jpg cat [ok])]  ACCEPTED, POOR TIME
                (
                    [(images[1], dog), (images[5], cat), (control_images[1][0], dog), (control_images[2][0], cat)],
                    'fedor',
                    10,
                ),
                #   alice [(4.jpg dog) (2_control.jpg dog [bad]) (1.jpg cat) (1_control.jpg cat [bad])] REJECTED
                (
                    [(images[4], dog), (control_images[2][0], dog), (images[1], cat), (control_images[1][0], cat)],
                    'alice',
                    39,
                ),
                #   paul   [(0_control.jpg cat [ok]) (3_control.jpg cat [bad]) (2.jpg cat) (0.jpg cat)]  ACCEPTED
                (
                    [(control_images[0][0], cat), (control_images[3][0], cat), (images[2], cat), (images[0], cat)],
                    'paul',
                    30,
                ),
            ]
        ],
        # result (BEFORE -> SUBMITTED -> ACCEPTED -> RESULT):
        #   0.jpg (cat|john) -> (cat|paul) -> (cat|paul) -> (cat|john, cat|paul), 0.66 conf
        #   1.jpg () -> (dog|fedor, cat|alice) -> (dog|fedor) -> (dog|fedor), 0.83 conf, but only one answer
        #   2.jpg (dog|alice) -> (cat|paul) -> (cat|paul) -> (dog|alice, cat|paul), 0.4 conf
        #   3.jpg (cat|john, cat|alice) -> () -> () -> (cat|john, cat|alice), 0.66 conf, dropped from 0.9, now uncertain
        #   4.jpg () -> (dog|martha, dog|alice) -> () -> ()
        #   5.jpg () -> (cat|martha, cat|fedor) -> (cat|fedor) -> (cat|fedor), 0.83 conf, but only one answer
        #
        #   martha, alice and fedor are blocked now
        #
        # iteration 2, duration=40
        [
            lib.create_classification_assignment(
                image_class_pairs, control_images, user_id=user_id, assignment_start=assignment_start, pool_id=pool_id
            )[0]
            for image_class_pairs, user_id in [
                #   paul  [(1_control.jpg dog [ok]) (1.jpg dog) (4.jpg cat) (2_control.jpg cat [ok])]  ACCEPTED
                (
                    [(control_images[1][0], dog), (images[1], dog), (images[4], cat), (control_images[2][0], cat)],
                    'paul',
                ),
                #   ivan  [(0_control.jpg dog [bad]) (2_control.jpg cat [ok]) (4.jpg dog) (2.jpg cat)] ACCEPTED
                (
                    [(control_images[0][0], dog), (control_images[2][0], cat), (images[4], dog), (images[2], cat)],
                    'ivan',
                ),
                #   paul  [(3.jpg dog) (5.jpg cat) (0_control.jpg cat [ok]) (3_control.jpg dog [ok])]  ACCEPTED
                (
                    [(images[3], dog), (images[5], dog), (control_images[0][0], cat), (control_images[3][0], dog)],
                    'paul',
                ),
                # incomplete assignment (we ran out of regular tasks)
                #   tom  [(1_control.jpg cat [bad]) (0.jpg cat) (3_control.jpg dog [ok])] ACCEPTED
                ([(control_images[1][0], cat), (images[0], cat), (control_images[3][0], dog)], 'tom'),
            ]
        ],
        # result (BEFORE -> SUBMITTED -> ACCEPTED -> RESULT):
        #   0.jpg (cat|john, cat|bob) -> (cat|tom) -> (cat|tom) -> (cat|john, cat|bob, cat|tom), 0.93 conf
        #   1.jpg (dog|fedor) -> (dog|paul) -> (dog|paul) -> (dog|fedor, dog|paul), 0.97 conf
        #   2.jpg (dog|alice, cat|paul) -> (cat|ivan) -> (cat|ivan) -> (dog|alice, cat|paul, cat|ivan), 0.83 conf
        #   3.jpg (cat|john, cat|alice) -> (dog|paul) -> (dog|paul) -> (cat|john, cat|alice, dog|paul), 0.59 conf
        #   4.jpg () -> (cat|paul, dog|ivan) -> (cat|paul, dog|ivan) -> (cat|paul, dog|ivan), 0.70 conf
        #   5.jpg (cat|fedor) -> (dog|paul) -> (dog|paul) -> (cat|fedor, dog|paul), 0.54 conf
        #
        #   tom is now blocked, because he did only 1 / 2 control correctly
        #   but this is a minor offence with short block time, and we assume it has passed between the iterations
        #
        # iteration 3, duration=40
        [
            lib.create_classification_assignment(
                image_class_pairs, control_images, user_id=user_id, assignment_start=assignment_start, pool_id=pool_id
            )[0]
            for image_class_pairs, user_id in [
                #   tom  [(2_control.jpg cat [ok]) (4.jpg cat) (3.jpg crow) (0_control.jpg crow [bad])] ACCEPTED
                (
                    [(control_images[2][0], cat), (images[4], cat), (images[3], crow), (control_images[0][0], crow)],
                    'tom',
                ),
                #   dylan  [(5.jpg crow) (0_control.jpg cat [ok]) (3_control.jpg dog [ok])]  ACCEPTED
                ([(images[5], crow), (control_images[0][0], cat), (control_images[3][0], dog)], 'dylan'),
            ]
        ],
        # result (BEFORE -> SUBMITTED -> ACCEPTED -> RESULT):
        #   0.jpg (cat|john, cat|bob, cat|tom) -> () -> () -> (cat|john, cat|bob, cat|tom), 0.93 conf
        #   1.jpg (dog|fedor, dog|paul) -> () -> () -> (dog|fedor, dog|paul), 0.97 conf
        #   2.jpg (dog|alice, cat|paul, cat|ivan) -> () -> () -> (dog|alice, cat|paul, cat|ivan), 0.83 conf
        #   3.jpg (cat|john, cat|alice, dog|paul) -> (crow|tom) -> (crow|tom) ->
        #                               (cat|john, cat|alice, dog|paul, crow|tom), 0.55 conf, but 4 answers already
        #   4.jpg (cat|paul, dog|ivan) -> (cat|tom) -> (cat|tom) -> (cat|paul, dog|ivan, cat|tom), 0.83 conf
        #   5.jpg (cat|fedor, dog|paul) -> (crow|dylan) -> (crow|dylan) -> (cat|fedor, dog|paul, crow|dylan), 0.36 conf
        #
        #
        #   tom is blocked now for 1 / 2 control tasks done correctly one more time
        #
        # iteration 4, duration=40
        [
            lib.create_classification_assignment(
                image_class_pairs,
                control_images,
                user_id=user_id,
                duration=datetime.timedelta(seconds=duration),
                assignment_start=assignment_start,
                pool_id=pool_id,
            )[0]
            for image_class_pairs, user_id, duration in [
                # assignment incompleteness is considered by prior filter, not rejected, but restricted
                #   theresa  [(2_control.jpg cat [ok]) (5.jpg dog) (1_control.jpg dog [ok])]  ACCEPTED
                ([(control_images[2][0], cat), (images[5], crow), (control_images[1][0], dog)], 'theresa', 4)
            ]
        ],
        # result (BEFORE -> SUBMITTED -> ACCEPTED -> RESULT):
        #   0.jpg (cat|john, cat|bob, cat|tom) -> () -> () -> (cat|john, cat|bob, cat|tom), 0.93 conf
        #   1.jpg (dog|fedor, dog|paul) -> () -> () -> (dog|fedor, dog|paul), 0.97 conf
        #   2.jpg (dog|alice, cat|paul, cat|ivan) -> () -> () -> (dog|alice, cat|paul, cat|ivan), 0.83 conf
        #   3.jpg (cat|john, cat|alice, dog|paul, crow|tom) -> () -> () ->
        #                               (cat|john, cat|alice, dog|paul, crow|tom), 0.55 conf, but 4 answers already
        #   4.jpg (cat|paul, dog|ivan, cat|tom) -> () -> () -> (cat|paul, dog|ivan, cat|tom), 0.83 conf
        #   5.jpg (cat|fedor, dog|paul, crow|dylan) -> (crow|theresa) -> (crow|theresa) ->
        #                               (cat|fedor, dog|paul, crow|dylan, crow|theresa), 0.85 conf
    ]
    i = 0
    for iteration_assignments in assignments:
        for assignment in iteration_assignments:
            assignment.id = f'{str(i)}_{assignment.user_id}'
            assignment.status = toloka.Assignment.SUBMITTED
            i += 1

    pool_input_objects = [(image,) for image in images]
    stub = TolokaClientStub(assignments)

    loop = classification_loop.ClassificationLoop(
        client=stub,  # noqa
        task_mapping=lib.image_classification_mapping,
        params=classification_loop.Params(
            aggregation_algorithm=classification.AggregationAlgorithm.MAX_LIKELIHOOD,
            overlap=classification_loop.DynamicOverlap(min_overlap=2, max_overlap=4, confidence=0.8),
            control=control_params,
            task_duration_function=duration.get_const_task_duration_function(task_duration_hint),
        ),
        lang='RU',
    )
    pool = loop.create_pool(control_objects=[((image,), (cls,)) for image, cls in control_images], pool_cfg=pool_cfg)
    loop.add_input_objects(pool.id, pool_input_objects)
    loop.loop(pool.id)
    results, worker_weights = loop.get_results(pool.id, pool_input_objects)

    assert results == [
        (
            {cat: approx(88 / 94), dog: approx(3 / 94), crow: approx(3 / 94)},
            [
                (cat, worker.Human(assignments[0][0])),
                (cat, worker.Human(assignments[1][3])),
                (cat, worker.Human(assignments[2][3])),
            ],
        ),
        (
            {cat: approx(3 / 226), dog: approx(220 / 226), crow: approx(3 / 226)},
            [(dog, worker.Human(assignments[1][1])), (dog, worker.Human(assignments[2][0]))],
        ),
        (
            {cat: approx(44 / 53), dog: approx(6 / 53), crow: approx(3 / 53)},
            [
                (dog, worker.Human(assignments[0][4])),
                (cat, worker.Human(assignments[1][3])),
                (cat, worker.Human(assignments[2][1])),
            ],
        ),
        (
            {cat: approx(30 / 100), dog: approx(55 / 100), crow: approx(15 / 100)},
            [
                (cat, worker.Human(assignments[0][0])),
                (cat, worker.Human(assignments[0][4])),
                (dog, worker.Human(assignments[2][2])),
                (crow, worker.Human(assignments[3][0])),
            ],
        ),
        (
            {cat: approx(44 / 53), dog: approx(6 / 53), crow: approx(3 / 53)},
            [
                (cat, worker.Human(assignments[2][0])),
                (dog, worker.Human(assignments[2][1])),
                (cat, worker.Human(assignments[3][0])),
            ],
        ),
        (
            {dog: approx(22 / 352), cat: approx(30 / 352), crow: approx(300 / 352)},
            [
                (cat, worker.Human(assignments[1][1])),
                (dog, worker.Human(assignments[2][2])),
                (crow, worker.Human(assignments[3][1])),
                (crow, worker.Human(assignments[4][0])),
            ],
        ),
    ]

    assert worker_weights == {
        'john': 1 / 2,
        'bob': 1 / 10,
        'fedor': 5 / 6,
        'mary': 1 / 6,
        'alice': 1 / 2,
        'martha': 5 / 6,
        'paul': 99 / 126,
        'ivan': 1 / 2,
        'tom': 1 / 2,
        'dylan': 5 / 6,
        'theresa': 5 / 6,
    }

    assert stub.calls[1:] == [  # skip pool creation
        (
            'create_tasks',
            (
                [
                    toloka.Task(
                        input_values={
                            'image_url': 'https://storage.net/0_control.jpg',
                            'id': "Image(url='https://storage.net/0_control.jpg')",
                        },
                        known_solutions=[
                            toloka.Task.KnownSolution(output_values={'choice': 'cat'}, correctness_weight=1)
                        ],
                        id='task 0',
                        infinite_overlap=True,
                        pool_id=pool_id,
                    ),
                    toloka.Task(
                        input_values={
                            'image_url': 'https://storage.net/1_control.jpg',
                            'id': "Image(url='https://storage.net/1_control.jpg')",
                        },
                        known_solutions=[
                            toloka.Task.KnownSolution(output_values={'choice': 'dog'}, correctness_weight=1)
                        ],
                        id='task 1',
                        infinite_overlap=True,
                        pool_id=pool_id,
                    ),
                    toloka.Task(
                        input_values={
                            'image_url': 'https://storage.net/2_control.jpg',
                            'id': "Image(url='https://storage.net/2_control.jpg')",
                        },
                        known_solutions=[
                            toloka.Task.KnownSolution(output_values={'choice': 'cat'}, correctness_weight=1)
                        ],
                        id='task 2',
                        infinite_overlap=True,
                        pool_id=pool_id,
                    ),
                    toloka.Task(
                        input_values={
                            'image_url': 'https://storage.net/3_control.jpg',
                            'id': "Image(url='https://storage.net/3_control.jpg')",
                        },
                        known_solutions=[
                            toloka.Task.KnownSolution(output_values={'choice': 'dog'}, correctness_weight=1)
                        ],
                        id='task 3',
                        infinite_overlap=True,
                        pool_id=pool_id,
                    ),
                ],
            ),
        ),
        (
            'create_tasks',
            (
                [
                    toloka.Task(
                        input_values={
                            'image_url': 'https://storage.net/0.jpg',
                            'id': "Image(url='https://storage.net/0.jpg')",
                        },
                        id='task 4',
                        pool_id=pool_id,
                    ),
                    toloka.Task(
                        input_values={
                            'image_url': 'https://storage.net/1.jpg',
                            'id': "Image(url='https://storage.net/1.jpg')",
                        },
                        id='task 5',
                        pool_id=pool_id,
                    ),
                    toloka.Task(
                        input_values={
                            'image_url': 'https://storage.net/2.jpg',
                            'id': "Image(url='https://storage.net/2.jpg')",
                        },
                        id='task 6',
                        pool_id=pool_id,
                    ),
                    toloka.Task(
                        input_values={
                            'image_url': 'https://storage.net/3.jpg',
                            'id': "Image(url='https://storage.net/3.jpg')",
                        },
                        id='task 7',
                        pool_id=pool_id,
                    ),
                    toloka.Task(
                        input_values={
                            'image_url': 'https://storage.net/4.jpg',
                            'id': "Image(url='https://storage.net/4.jpg')",
                        },
                        id='task 8',
                        pool_id=pool_id,
                    ),
                    toloka.Task(
                        input_values={
                            'image_url': 'https://storage.net/5.jpg',
                            'id': "Image(url='https://storage.net/5.jpg')",
                        },
                        id='task 9',
                        pool_id=pool_id,
                    ),
                ],
            ),
        ),
        ('open_pool', (pool_id,)),
        # Apply prior filter by time first
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    private_comment='Fast submits, <= 0.1',
                    user_id='john',
                    pool_id=pool_id,
                    will_expire=assignment_start + datetime.timedelta(seconds=3) + datetime.timedelta(days=1),
                ),
            ),
        ),
        (
            'patch_assignment',
            (
                '5_john',
                toloka.assignment.AssignmentPatch(
                    status=toloka.Assignment.REJECTED, public_comment='Мало правильных решений'
                ),
            ),
        ),
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    private_comment='Control tasks: [1, 2) done correctly',
                    user_id='john',
                    pool_id=pool_id,
                    will_expire=assignment_start + datetime.timedelta(seconds=35) + datetime.timedelta(hours=1),
                ),
            ),
        ),
        (
            'patch_assignment',
            ('0_john', toloka.assignment.AssignmentPatch(status=toloka.Assignment.ACCEPTED, public_comment='')),
        ),
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    private_comment='Control tasks: [0, 1) done correctly',
                    user_id='bob',
                    pool_id=pool_id,
                    will_expire=assignment_start + datetime.timedelta(seconds=31) + datetime.timedelta(hours=8),
                ),
            ),
        ),
        (
            'patch_assignment',
            (
                '1_bob',
                toloka.assignment.AssignmentPatch(
                    status=toloka.Assignment.REJECTED,
                    public_comment='Проверьте задания c номерами: 1, 4. Подробнее о проверке заданий и подаче апелляций '
                    'читайте в инструкции проекта.',
                ),
            ),
        ),
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    private_comment='Control tasks: [0, 1) done correctly',
                    user_id='mary',
                    pool_id=pool_id,
                    will_expire=assignment_start + datetime.timedelta(seconds=39) + datetime.timedelta(hours=8),
                ),
            ),
        ),
        (
            'patch_assignment',
            (
                '2_mary',
                toloka.assignment.AssignmentPatch(
                    status=toloka.Assignment.REJECTED,
                    public_comment='Проверьте задания c номерами: 3, 4. Подробнее о проверке заданий и подаче апелляций '
                    'читайте в инструкции проекта.',
                ),
            ),
        ),
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    private_comment='Control tasks: [0, 1) done correctly',
                    user_id='bob',
                    pool_id=pool_id,
                    will_expire=assignment_start + datetime.timedelta(seconds=20) + datetime.timedelta(hours=8),
                ),
            ),
        ),
        (
            'patch_assignment',
            (
                '3_bob',
                toloka.assignment.AssignmentPatch(
                    status=toloka.Assignment.REJECTED,
                    public_comment='Проверьте задания c номерами: 2, 3. Подробнее о проверке заданий и подаче апелляций '
                    'читайте в инструкции проекта.',
                ),
            ),
        ),
        (
            'patch_assignment',
            ('4_alice', toloka.assignment.AssignmentPatch(status=toloka.Assignment.ACCEPTED, public_comment='')),
        ),
        ('patch_task_overlap_or_min', ('task 4', toloka.task.TaskOverlapPatch(overlap=3))),
        ('patch_task_overlap_or_min', ('task 6', toloka.task.TaskOverlapPatch(overlap=3))),
        ('patch_task_overlap_or_min', ('task 5', toloka.task.TaskOverlapPatch(overlap=4))),
        ('patch_task_overlap_or_min', ('task 9', toloka.task.TaskOverlapPatch(overlap=4))),
        ('patch_task_overlap_or_min', ('task 8', toloka.task.TaskOverlapPatch(overlap=4))),
        ('open_pool', (pool_id,)),
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    private_comment='Fast submits, <= 0.1',
                    user_id='martha',
                    pool_id=pool_id,
                    will_expire=assignment_start + datetime.timedelta(seconds=2) + datetime.timedelta(days=1),
                ),
            ),
        ),
        (
            'patch_assignment',
            (
                '6_martha',
                toloka.assignment.AssignmentPatch(
                    status=toloka.Assignment.REJECTED, public_comment='Мало правильных решений'
                ),
            ),
        ),
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    private_comment='Fast submits, 0.1 < time <= 0.3',
                    user_id='fedor',
                    pool_id=pool_id,
                    will_expire=assignment_start + datetime.timedelta(seconds=10) + datetime.timedelta(hours=8),
                ),
            ),
        ),
        (
            'patch_assignment',
            ('7_fedor', toloka.assignment.AssignmentPatch(status=toloka.Assignment.ACCEPTED, public_comment='')),
        ),
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    private_comment='Control tasks: [0, 1) done correctly',
                    user_id='alice',
                    pool_id=pool_id,
                    will_expire=assignment_start + datetime.timedelta(seconds=39) + datetime.timedelta(hours=8),
                ),
            ),
        ),
        (
            'patch_assignment',
            (
                '8_alice',
                toloka.assignment.AssignmentPatch(
                    status=toloka.Assignment.REJECTED,
                    public_comment='Проверьте задания c номерами: 2, 4. Подробнее о проверке заданий и подаче апелляций '
                    'читайте в инструкции проекта.',
                ),
            ),
        ),
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    private_comment='Control tasks: [1, 2) done correctly',
                    user_id='paul',
                    pool_id=pool_id,
                    will_expire=assignment_start + datetime.timedelta(seconds=30) + datetime.timedelta(hours=1),
                ),
            ),
        ),
        (
            'patch_assignment',
            ('9_paul', toloka.assignment.AssignmentPatch(status=toloka.Assignment.ACCEPTED, public_comment='')),
        ),
        ('patch_task_overlap_or_min', ('task 4', toloka.task.TaskOverlapPatch(overlap=4))),
        ('patch_task_overlap_or_min', ('task 7', toloka.task.TaskOverlapPatch(overlap=3))),
        ('patch_task_overlap_or_min', ('task 6', toloka.task.TaskOverlapPatch(overlap=4))),
        ('patch_task_overlap_or_min', ('task 5', toloka.task.TaskOverlapPatch(overlap=5))),
        ('patch_task_overlap_or_min', ('task 9', toloka.task.TaskOverlapPatch(overlap=5))),
        ('patch_task_overlap_or_min', ('task 8', toloka.task.TaskOverlapPatch(overlap=6))),
        ('open_pool', (pool_id,)),
        (
            'patch_assignment',
            ('10_paul', toloka.assignment.AssignmentPatch(status=toloka.Assignment.ACCEPTED, public_comment='')),
        ),
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    private_comment='Control tasks: [1, 2) done correctly',
                    user_id='ivan',
                    pool_id=pool_id,
                    will_expire=assignment_start + datetime.timedelta(seconds=40) + datetime.timedelta(hours=1),
                ),
            ),
        ),
        (
            'patch_assignment',
            ('11_ivan', toloka.assignment.AssignmentPatch(status=toloka.Assignment.ACCEPTED, public_comment='')),
        ),
        (
            'patch_assignment',
            ('12_paul', toloka.assignment.AssignmentPatch(status=toloka.Assignment.ACCEPTED, public_comment='')),
        ),
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    private_comment='Control tasks: [1, 2) done correctly',
                    user_id='tom',
                    pool_id=pool_id,
                    will_expire=assignment_start + datetime.timedelta(seconds=40) + datetime.timedelta(hours=1),
                ),
            ),
        ),
        (
            'patch_assignment',
            ('13_tom', toloka.assignment.AssignmentPatch(status=toloka.Assignment.ACCEPTED, public_comment='')),
        ),
        ('patch_task_overlap_or_min', ('task 7', toloka.task.TaskOverlapPatch(overlap=4))),
        ('patch_task_overlap_or_min', ('task 9', toloka.task.TaskOverlapPatch(overlap=6))),
        ('patch_task_overlap_or_min', ('task 8', toloka.task.TaskOverlapPatch(overlap=7))),
        ('open_pool', (pool_id,)),
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    private_comment='Control tasks: [1, 2) done correctly',
                    user_id='tom',
                    pool_id=pool_id,
                    will_expire=assignment_start + datetime.timedelta(seconds=40) + datetime.timedelta(hours=1),
                ),
            ),
        ),
        (
            'patch_assignment',
            ('14_tom', toloka.assignment.AssignmentPatch(status=toloka.Assignment.ACCEPTED, public_comment='')),
        ),
        (
            'patch_assignment',
            ('15_dylan', toloka.assignment.AssignmentPatch(status=toloka.Assignment.ACCEPTED, public_comment='')),
        ),
        ('patch_task_overlap_or_min', ('task 9', toloka.task.TaskOverlapPatch(overlap=7))),
        ('open_pool', (pool_id,)),
        (
            'set_user_restriction',
            (
                toloka.user_restriction.PoolUserRestriction(
                    private_comment='Fast submits, 0.1 < time <= 0.3',
                    user_id='theresa',
                    pool_id=pool_id,
                    will_expire=assignment_start + datetime.timedelta(seconds=4) + datetime.timedelta(hours=8),
                ),
            ),
        ),
        (
            'patch_assignment',
            ('16_theresa', toloka.assignment.AssignmentPatch(status=toloka.Assignment.ACCEPTED, public_comment='')),
        ),
    ]


class TestTaskIdToOverlapIncrease:
    images = [Image(url=f'https://storage.net/{i}.jpg') for i in range(4)]
    control_images = [
        (Image(url=f'https://storage.net/{i}_control.jpg'), lib.cat if i % 2 == 0 else lib.dog) for i in range(4)
    ]

    class TolokaClientStub:
        def __init__(self, assignments: List[toloka.Assignment]):
            self.assignments = assignments
            self.tasks = [
                lib.image_classification_mapping.to_task((image,)) for image in TestTaskIdToOverlapIncrease.images
            ]
            for task in self.tasks:
                task.id = task.input_values[mapping.TASK_ID_FIELD]

        def get_assignments(
            self,
            status: Union[toloka.Assignment.Status, List[toloka.Assignment.Status]],
            pool_id: str,
        ):
            if not isinstance(status, list):
                status = [status]
            return [assignment for assignment in self.assignments if assignment.status in status]

        def get_tasks(self, *args, **kwargs):
            return self.tasks

    def get_task_id_overlaps(self, image_index_to_overlap: Dict[int, int]) -> Dict[mapping.TaskID, int]:
        return {mapping.TaskID((self.images[index],)): overlap for index, overlap in image_index_to_overlap.items()}

    def test_static_overlap(self):
        cat = lib.cat
        images = self.images

        def gen_assignment_objects(*args) -> List[Tuple[Image, lib.ImageClass]]:
            return [(images[i], cat) for i in args]

        assignments_objects = [
            gen_assignment_objects(0, 1, 3),
            gen_assignment_objects(1, 2, 3),
            gen_assignment_objects(0, 2, 3),
            gen_assignment_objects(0, 1, 2),
        ]

        for objects, statuses, overlap, expected_task_id_to_overlap_increase in [
            (assignments_objects, [toloka.Assignment.ACCEPTED] * 4, 3, {}),
            (
                assignments_objects,
                [toloka.Assignment.REJECTED] * 2 + [toloka.Assignment.ACCEPTED] * 2,
                3,
                self.get_task_id_overlaps({0: 1, 1: 2, 2: 1, 3: 2}),
            ),
            (
                assignments_objects,
                [toloka.Assignment.ACCEPTED] * 2 + [toloka.Assignment.REJECTED] * 2,
                3,
                self.get_task_id_overlaps({0: 2, 1: 1, 2: 2, 3: 1}),
            ),
            (
                assignments_objects[:2],
                [toloka.Assignment.ACCEPTED] * 2,
                3,
                self.get_task_id_overlaps({0: 2, 1: 1, 2: 2, 3: 1}),
            ),
            (
                assignments_objects,
                [toloka.Assignment.ACCEPTED] * 4,
                4,
                self.get_task_id_overlaps({0: 1, 1: 1, 2: 1, 3: 1}),
            ),
            (
                assignments_objects,
                [toloka.Assignment.REJECTED] * 2 + [toloka.Assignment.ACCEPTED] * 2,
                4,
                self.get_task_id_overlaps({0: 2, 1: 3, 2: 2, 3: 3}),
            ),
        ]:
            assignments = [
                lib.create_classification_assignment(assignment_objects, [])[0] for assignment_objects in objects
            ]
            for assignment, status in zip(assignments, statuses):
                assignment.status = status
            stub = self.TolokaClientStub(assignments)
            loop = classification_loop.ClassificationLoop(
                client=stub,  # noqa
                task_mapping=lib.image_classification_mapping,
                params=classification_loop.Params(
                    aggregation_algorithm=classification.AggregationAlgorithm.MAJORITY_VOTE,
                    overlap=classification_loop.StaticOverlap(overlap),
                    control=control.Control(rules=control.RuleBuilder().add_static_reward(threshold=0.5).build()),
                    task_duration_function=None,  # noqa
                ),
                lang='RU',
            )
            assert loop.get_task_id_to_overlap_increase('fake') == expected_task_id_to_overlap_increase

    def test_dynamic_overlap(self):
        dog, cat, crow = lib.ImageClass.possible_instances()
        control_images = self.control_images
        images = self.images
        all_assignments = [
            lib.create_classification_assignment(image_class_pairs, control_images, user_id=user_id)[0]
            for image_class_pairs, user_id in [
                #   john  [(0.jpg cat) (1_control.jpg dog [ok]) (3.jpg dog) (2_control.jpg cat [ok])]
                (
                    [(images[0], cat), (control_images[1][0], dog), (images[3], dog), (control_images[2][0], cat)],
                    'john',
                ),
                #   john   [(0_control.jpg cat [ok]) (2.jpg cat) (1.jpg dog) (2_control.jpg cat [ok])]
                (
                    [(control_images[0][0], cat), (images[2], cat), (images[1], dog), (control_images[2][0], cat)],
                    'john',
                ),
                #   mary  [(0.jpg cat) (2.jpg dog) (1_control.jpg dog [ok]) (3_control.jpg dog [ok])]
                (
                    [(images[0], cat), (images[2], dog), (control_images[1][0], dog), (control_images[3][0], dog)],
                    'mary',
                ),
                #   bob   [(0.jpg dog) (3_control.jpg dog [ok]) (1_control.jpg dog [ok]) (2.jpg cat)]
                ([(images[0], dog), (control_images[3][0], dog), (control_images[1][0], dog), (images[2], cat)], 'bob'),
                #   alice [(2.jpg dog) (3.jpg cat) (2_control.jpg dog [bad]) (0_control.jpg dog [bad])]
                (
                    [(images[2], dog), (images[3], cat), (control_images[2][0], dog), (control_images[0][0], dog)],
                    'alice',
                ),
                #   alice  [(2_control.jpg cat [ok]) (0.jpg dog) (1.jpg cat) (0_control.jpg cat [ok])]
                (
                    [(control_images[2][0], cat), (images[0], dog), (images[1], cat), (control_images[0][0], cat)],
                    'alice',
                ),
            ]
        ]

        accepted, rejected = toloka.Assignment.ACCEPTED, toloka.Assignment.REJECTED
        for assignment_indices, statuses, min_overlap, max_overlap, conf, expected_task_id_to_overlap_increase in [
            [
                (0, 1),
                (accepted, rejected),
                3,
                4,
                1.0,  # residual, tasks 0 & 3 have 1 accepted solution each
                self.get_task_id_overlaps({0: 2, 1: 3, 2: 3, 3: 2}),
            ],
            # 0 - 4 attempts already, no more attempts can be done
            # 1 - 2 attempts, 0.85 confidence, not enough, need 1 more attempt at least
            # 2 - 3 attempts, 0.94 confidence, enough, no more attempts needed
            # 3 - 1 attempt only, need 1 more for min_overlap
            [
                (0, 1, 2, 3, 4, 5),
                (accepted, accepted, accepted, accepted, rejected, accepted),
                2,
                4,
                0.9,
                self.get_task_id_overlaps({1: 1, 3: 1}),
            ],
        ]:
            assignments = [all_assignments[i] for i in assignment_indices]
            for assignment, status in zip(assignments, statuses):
                assignment.status = status

            stub = self.TolokaClientStub(assignments)
            loop = classification_loop.ClassificationLoop(
                client=stub,  # noqa
                task_mapping=lib.image_classification_mapping,
                params=classification_loop.Params(
                    aggregation_algorithm=classification.AggregationAlgorithm.MAX_LIKELIHOOD,
                    overlap=classification_loop.DynamicOverlap(
                        min_overlap=min_overlap, max_overlap=max_overlap, confidence=conf
                    ),
                    control=control.Control(rules=control.RuleBuilder().add_static_reward(threshold=0.5).build()),
                    task_duration_function=None,  # noqa
                ),
                lang='RU',
            )
            assert loop.get_task_id_to_overlap_increase('fake') == expected_task_id_to_overlap_increase


class TolokaClientStub(lib.TolokaClientCallRecorderStub):
    def __init__(self, tasks: List[toloka.Task]):
        self.tasks = tasks
        super(TolokaClientStub, self).__init__()

    def get_tasks(self, request: toloka.search_requests.TaskSearchRequest):
        return self.tasks


def test_re_markup_not_finalized_objects():
    audios = [(Audio(url=f'https://storage.net/{i}.wav'),) for i in range(6)]

    def create_assignment(indexes: Iterable[int]) -> mapping.AssignmentSolutions:
        # text is not important here
        return lib.create_markup_assignment([(audios[i][0], Text(text='')) for i in indexes])

    assignments = [
        # [0] #1 attempt BAD, [1] #1 attempt BAD, [2] #1 attempt BAD, [3] #1 attempt OK
        create_assignment([0, 1, 2, 3]),
        # [0] #2 attempt BAD, [1] #2 attempt OK, [4] #1 attempt BAD, [5] #1 attempt OK
        create_assignment([0, 1, 4, 5]),
    ]

    tasks = []
    for i, audio in enumerate(audios):
        task = lib.audio_transcript_mapping.to_task(audio)
        task.id = f'task-{i}'
        tasks.append(task)

    for task_id_to_overlap_increase, expected_objects_to_re_markup, expected_toloka_calls in (
        (
            {task_id: 1 for task_id in [lib.audio_transcript_mapping.task_id(audios[i]) for i in (0, 2, 4)]},
            3,
            [
                ('patch_task_overlap_or_min', ('task-0', toloka.task.TaskOverlapPatch(overlap=3))),
                ('patch_task_overlap_or_min', ('task-2', toloka.task.TaskOverlapPatch(overlap=2))),
                ('patch_task_overlap_or_min', ('task-4', toloka.task.TaskOverlapPatch(overlap=2))),
                ('open_pool', ('fake',)),
            ],
        ),
        (
            {},
            0,
            [],
        ),
    ):
        stub = TolokaClientStub(tasks)
        tasks_to_rework = classification_loop.rework_not_finalized_tasks(
            client=stub,  # noqa
            assignments=assignments,
            task_mapping=lib.audio_transcript_mapping,
            task_id_to_overlap_increase=task_id_to_overlap_increase,
            pool_id='fake',
        )

        assert tasks_to_rework == expected_objects_to_re_markup
        assert stub.calls == expected_toloka_calls


def test_calculate_label_probas():
    dog, cat, crow = lib.dog, lib.cat, lib.crow
    images = [Image(url=f'https://storage.net/{i}.jpg') for i in range(4)]
    control_images = [(Image(url=f'https://storage.net/{i}_control.jpg'), cat if i % 2 == 0 else dog) for i in range(4)]
    task_mapping = lib.image_classification_mapping
    assignment_evaluation_strategy = evaluation.ControlTasksAssignmentAccuracyEvaluationStrategy(task_mapping)
    assignment_solutions = [
        lib.create_classification_assignment(image_class_pairs, control_images, user_id=user_id, status=status)
        for image_class_pairs, user_id, status in [
            ([(images[0], cat), (control_images[0][0], cat), (images[1], dog)], 'john', toloka.Assignment.ACCEPTED),
            ([(images[2], cat), (images[3], crow), (control_images[1][0], crow)], 'kate', toloka.Assignment.REJECTED),
            ([(control_images[2][0], crow), (images[2], dog), (images[0], cat)], 'nolan', toloka.Assignment.REJECTED),
            ([(images[1], dog), (control_images[3][0], dog), (images[3], dog)], 'nolan', toloka.Assignment.ACCEPTED),
        ]
    ]

    # we collect worker weights from both accepted and rejected assignments
    # however, for label aggregation we only use answers from accepted assignments
    # images[0] - [(cat|john)]
    # images[1] - [(dog|john), (dog|nolan)]
    # images[2] - [] - no answers at all
    # images[3] - [(dog|nolan)]

    tasks = []
    task_ids = []
    for i, image in enumerate(images):
        task = task_mapping.to_task((image,))
        task.id = f'task-{i}'
        tasks.append(task)
        task_id = mapping.TaskID((image,))
        task_ids.append(task_id)

    stub = TolokaClientStub(tasks)
    expected = {task_ids[0]: (cat, approx(0.75)), task_ids[1]: (dog, approx(6 / 7)), task_ids[3]: (dog, approx(0.5))}

    assert (
        classification_loop.calculate_label_probas(
            client=stub,  # noqa
            task_mapping=task_mapping,
            assignment_evaluation_strategy=assignment_evaluation_strategy,
            aggregation_algorithm=classification.AggregationAlgorithm.MAX_LIKELIHOOD,
            assignment_solutions=assignment_solutions,
            pool_id='fake',
        )
        == expected
    )



from datetime import timedelta
from mock import patch
from pytest import approx, raises

from crowdom import base, classification_loop, evaluation, mapping, objects, pricing
from . import lib

stub_mapping = task_mapping = mapping.TaskMapping(
    input_mapping=(),
    output_mapping=(
        mapping.ObjectMapping(obj_meta=base.ObjectMeta(lib.ImageClass), obj_task_fields=(('value', 'choice'),)),
    ),
)

annotation_stub_mapping = mapping.TaskMapping(
    input_mapping=(
        mapping.ObjectMapping(obj_meta=base.ObjectMeta(objects.Audio), obj_task_fields=(('url', 'audio'),)),
    ),
    output_mapping=(
        mapping.ObjectMapping(obj_meta=base.ObjectMeta(objects.Text), obj_task_fields=(('text', 'transcript'),)),
    ),
)


class TestClassification:
    def test_pricing(self):
        options = pricing.get_pricing_options(
            task_duration_hint=timedelta(seconds=9),
            pricing_strategy=pricing.StaticPricingStrategy(),
            lang='RU',
            task_mapping=stub_mapping,
        )

        assert options[0] == pricing.PoolPricingConfigWithCalculatedProperties(
            config=pricing.PoolPricingConfig(assignment_price=0.006, real_tasks_count=3, control_tasks_count=1),
            price_per_hour=0.006 * (60 * 60 / ((3 + 1) * 9)),
            robustness=1 - 1 / 3,
            control_tasks_ratio=1 / (3 + 1),
            assignment_duration=timedelta(seconds=9 * (3 + 1)),
        )

        assert pricing.choose_default_option(options) == pricing.PoolPricingConfigWithCalculatedProperties(
            config=pricing.PoolPricingConfig(assignment_price=0.024, real_tasks_count=14, control_tasks_count=2),
            price_per_hour=0.024 * (60 * 60 / ((14 + 2) * 9)),
            robustness=1 - (1 / 3) ** 2,
            control_tasks_ratio=2 / (14 + 2),
            assignment_duration=timedelta(seconds=9 * (14 + 2)),
        )

        assert options[-1] == pricing.PoolPricingConfigWithCalculatedProperties(
            config=pricing.PoolPricingConfig(assignment_price=0.043, real_tasks_count=19, control_tasks_count=10),
            price_per_hour=0.043 * (60 * 60 / ((19 + 10) * 9)),
            robustness=1 - (1 / 3) ** 10,
            control_tasks_ratio=10 / (19 + 10),
            assignment_duration=timedelta(seconds=9 * (19 + 10)),
        )

    def test_pricing_with_adjustment(self):
        options = pricing.get_pricing_options(
            task_duration_hint=timedelta(seconds=9),
            pricing_strategy=pricing.StaticPricingStrategy(),
            lang='RU',
            task_mapping=stub_mapping,
            correct_control_task_ratio_for_acceptance=0.8,
        )

        assert options[0] == pricing.PoolPricingConfigWithCalculatedProperties(
            config=pricing.PoolPricingConfig(assignment_price=0.006, real_tasks_count=3, control_tasks_count=1),
            price_per_hour=0.006 * (60 * 60 / ((3 + 1) * 9)),
            robustness=1 - 1 / 3,
            control_tasks_ratio=1 / (3 + 1),
            assignment_duration=timedelta(seconds=9 * (3 + 1)),
        )

        assert (
            pricing.choose_default_option(options)
            == pricing.PoolPricingConfigWithCalculatedProperties(
                config=pricing.PoolPricingConfig(assignment_price=0.024, real_tasks_count=14, control_tasks_count=2),
                price_per_hour=0.024 * (60 * 60 / ((14 + 2) * 9)),
                robustness=1 - ((1 / 3) ** 2 * (1 - 1 / 3) ** 0),
                control_tasks_ratio=2 / (14 + 2),
                assignment_duration=timedelta(seconds=9 * (14 + 2)),
            )
            == pricing.calculate_properties_for_pricing_config(
                config=pricing.PoolPricingConfig(assignment_price=0.024, real_tasks_count=14, control_tasks_count=2),
                task_duration_hint=timedelta(seconds=9),
                correct_control_task_ratio_for_acceptance=0.8,
                task_mapping=stub_mapping,
            )
        )

        assert options[-1] == pricing.PoolPricingConfigWithCalculatedProperties(
            config=pricing.PoolPricingConfig(assignment_price=0.043, real_tasks_count=19, control_tasks_count=10),
            price_per_hour=0.043 * (60 * 60 / ((19 + 10) * 9)),
            robustness=1
            - (
                ((1 / 3) ** 8 * (1 - 1 / 3) ** 2) * 5 * 9
                + ((1 / 3) ** 9 * (1 - 1 / 3) ** 1) * 10  # C_10_8 = 10! / (8! * 2!)
                + ((1 / 3) ** 10 * (1 - 1 / 3) ** 0) * 1
            ),
            control_tasks_ratio=10 / (19 + 10),
            assignment_duration=timedelta(seconds=9 * (19 + 10)),
        )


class TestExperts:
    def test_pricing_inhouse(self):
        options = pricing.get_expert_pricing_options(task_duration_hint=timedelta(seconds=9), task_mapping=stub_mapping)

        assert options[0] == pricing.PoolPricingConfig(
            assignment_price=0.005, real_tasks_count=4, control_tasks_count=0
        )
        assert pricing.choose_default_expert_option(options, None) == pricing.PoolPricingConfig(
            assignment_price=0.005, real_tasks_count=29, control_tasks_count=0
        )
        assert options[-1] == pricing.PoolPricingConfig(
            assignment_price=0.005, real_tasks_count=29, control_tasks_count=0
        )

    def test_pricing_external(self):
        options = pricing.get_expert_pricing_options(
            task_duration_hint=timedelta(seconds=9), task_mapping=stub_mapping, avg_price_per_hour=2.5
        )

        assert options[0] == pricing.PoolPricingConfig(
            assignment_price=0.025, real_tasks_count=4, control_tasks_count=0
        )
        assert pricing.choose_default_expert_option(options, 2.5) == pricing.PoolPricingConfig(
            assignment_price=0.025, real_tasks_count=4, control_tasks_count=0
        )
        assert options[-1] == pricing.PoolPricingConfig(
            assignment_price=0.181, real_tasks_count=29, control_tasks_count=0
        )


class TestTraining:
    def test_get_options_not_enough_tasks(self):
        with raises(AssertionError) as e:
            pricing.get_training_options(task_duration_hint=timedelta(seconds=10), available_task_count=1)
        assert (
            str(e.value)
            == 'You do not have enough tasks for this training. Expected at least 18 tasks for this parameters, got 1'
        )

    def test_get_options_wrong_duration(self):
        with raises(AssertionError) as e:
            pricing.get_training_options(
                task_duration_hint=timedelta(seconds=10), available_task_count=1000, training_time=timedelta(hours=3)
            )
        assert (
            str(e.value)
            == 'You are trying to create either too short or too long training. Expected from 2 to 30 tasks, got 1080'
        )

        with raises(AssertionError) as e:
            pricing.get_training_options(
                task_duration_hint=timedelta(seconds=10), available_task_count=1000, training_time=timedelta(seconds=10)
            )
        assert (
            str(e.value)
            == 'You are trying to create either too short or too long training. Expected from 2 to 30 tasks, got 1'
        )

    def test_get_options(self):
        assert pricing.get_training_options(task_duration_hint=timedelta(seconds=10), available_task_count=1000) == [
            pricing.TrainingConfig(18, 1, timedelta(seconds=1800)),
            pricing.TrainingConfig(9, 2, timedelta(seconds=900)),
            pricing.TrainingConfig(6, 3, timedelta(seconds=600)),
            pricing.TrainingConfig(5, 4, timedelta(seconds=500)),
            pricing.TrainingConfig(4, 5, timedelta(seconds=400)),
        ]

    def test_choose_default_option(self):
        options = [
            pricing.TrainingConfig(20, 1, timedelta(seconds=600)),
            pricing.TrainingConfig(10, 2, timedelta(seconds=300)),
            pricing.TrainingConfig(5, 4, timedelta(seconds=150)),
        ]
        assert pricing.choose_default_training_option(options) == options[1]
        assert pricing.choose_default_training_option(options[:2]) == options[1]

    def test_choose_default_option_empty(self):
        with raises(AssertionError) as e:
            pricing.choose_default_training_option([])

        assert str(e.value) == 'No available options'


class TestAnnotation:
    def test_get_options(self):
        options = pricing.get_annotation_pricing_options(
            task_duration_hint=timedelta(seconds=10),
            pricing_strategy=pricing.StaticPricingStrategy(),
            lang='EN',
            task_mapping=annotation_stub_mapping,
        )
        assert len(options) == 27

        assert options[0] == pricing.PoolPricingConfig(
            assignment_price=0.008, real_tasks_count=3, control_tasks_count=0
        )
        assert options[-1] == pricing.PoolPricingConfig(
            assignment_price=0.081, real_tasks_count=29, control_tasks_count=0
        )

    def test_choose_default_option(self):
        options = [
            pricing.PoolPricingConfig(20, 1, timedelta(seconds=600)),
            pricing.PoolPricingConfig(10, 2, timedelta(seconds=300)),
            pricing.PoolPricingConfig(5, 4, timedelta(seconds=150)),
        ]
        assert pricing.choose_default_annotation_option(options) == options[1]
        assert pricing.choose_default_annotation_option(options[:2]) == options[1]

    def test_choose_default_option_empty(self):
        with raises(AssertionError) as e:
            pricing.choose_default_annotation_option([])

        assert str(e.value) == 'No available options'


class TestPoolPriceFormula:
    def test_simple(self):
        formula = pricing.PoolPriceFormula(
            input_objects_count=48,
            config=pricing.PoolPricingConfig(assignment_price=0.06, real_tasks_count=12, control_tasks_count=4),
            overlap=classification_loop.StaticOverlap(overlap=3),
        )
        assert formula.min_total_price_clear == formula.max_total_price_clear == approx(0.936)
        assert (
            formula.clear_formula()
            == r'TotalPrice_{clear} = TaskCount * PricePerTask_\$ * Overlap * (1 + TolokaCommission) = '
            r'48 * 0.0050\$ * 3 * 1.3 = 0.94\$.'
        )

        assert formula.min_total_price_precise == formula.max_total_price_clear == approx(0.936)
        assert (
            formula.precise_formula()
            == r'TotalPrice_{precise} = AssignmentCount * PricePerAssignment_\$ * Overlap = \left \lceil '
            r'\frac {TaskCount} {TasksOnAssignment} \right \rceil * (PricePerAssignment_\$ + '
            r'max(PricePerAssignment_\$ * TolokaCommission, MinTolokaCommission_\$) * Overlap = '
            r'\lceil 48 / 12 \rceil * (0.06\$ + max(0.06\$ * 0.3, 0.001\$) * 3 = 4 * 0.078 * 3 = 0.94\$.'
        )

    def test_incomplete_last_assignment(self):
        formula = pricing.PoolPriceFormula(
            input_objects_count=50,
            config=pricing.PoolPricingConfig(assignment_price=0.06, real_tasks_count=12, control_tasks_count=4),
            overlap=classification_loop.StaticOverlap(overlap=3),
        )
        assert formula.min_total_price_clear == formula.max_total_price_clear == approx(0.975)
        assert (
            formula.clear_formula()
            == r'TotalPrice_{clear} = TaskCount * PricePerTask_\$ * Overlap * (1 + TolokaCommission) = '
            r'50 * 0.0050\$ * 3 * 1.3 = 0.98\$.'
        )

        assert formula.min_total_price_precise == formula.max_total_price_precise == approx(1.17)
        assert (
            formula.precise_formula()
            == r'TotalPrice_{precise} = AssignmentCount * PricePerAssignment_\$ * Overlap = \left \lceil '
            r'\frac {TaskCount} {TasksOnAssignment} \right \rceil * (PricePerAssignment_\$ + '
            r'max(PricePerAssignment_\$ * TolokaCommission, MinTolokaCommission_\$) * Overlap = '
            r'\lceil 50 / 12 \rceil * (0.06\$ + max(0.06\$ * 0.3, 0.001\$) * 3 = 5 * 0.078 * 3 = 1.17\$.'
        )

    @patch('crowdom.pricing.min_toloka_commission', 0.005)
    def test_min_commission(self):
        formula = pricing.PoolPriceFormula(
            input_objects_count=48,
            config=pricing.PoolPricingConfig(assignment_price=0.01, real_tasks_count=12, control_tasks_count=4),
            overlap=classification_loop.StaticOverlap(overlap=3),
        )
        assert formula.min_total_price_clear == formula.max_total_price_clear == approx(0.156)
        assert (
            formula.clear_formula()
            == r'TotalPrice_{clear} = TaskCount * PricePerTask_\$ * Overlap * (1 + TolokaCommission) = '
            r'48 * 0.0008\$ * 3 * 1.3 = 0.16\$.'
        )

        assert formula.min_total_price_precise == formula.max_total_price_precise == approx(0.18)
        assert (
            formula.precise_formula()
            == r'TotalPrice_{precise} = AssignmentCount * PricePerAssignment_\$ * Overlap = \left \lceil '
            r'\frac {TaskCount} {TasksOnAssignment} \right \rceil * (PricePerAssignment_\$ + '
            r'max(PricePerAssignment_\$ * TolokaCommission, MinTolokaCommission_\$) * Overlap = '
            r'\lceil 48 / 12 \rceil * (0.01\$ + max(0.01\$ * 0.3, 0.005\$) * 3 = 4 * 0.015 * 3 = 0.18\$.'
        )

    def test_dynamic_overlap(self):
        formula = pricing.PoolPriceFormula(
            input_objects_count=50,
            config=pricing.PoolPricingConfig(assignment_price=0.06, real_tasks_count=12, control_tasks_count=4),
            overlap=classification_loop.DynamicOverlap(min_overlap=2, max_overlap=4, confidence=0.5),
        )

        assert (formula.min_total_price_clear, formula.max_total_price_clear) == (approx(0.65), approx(1.3))
        assert (formula.min_total_price_precise, formula.max_total_price_precise) == (approx(0.78), approx(1.56))

        assert (
            formula.clear_formula()
            == r'TotalPrice_{clear} = TaskCount * PricePerTask_\$ * [MinOverlap \dots MaxOverlap] * '
            r'(1 + TolokaCommission) = 50 * 0.0050\$ * [2 \dots 4] * 1.3 = 0.65\$ \dots 1.30\$.'
        )

        assert (
            formula.precise_formula()
            == r'TotalPrice_{precise} = AssignmentCount * PricePerAssignment_\$ * [MinOverlap \dots MaxOverlap] = '
            r'\left \lceil \frac {TaskCount} {TasksOnAssignment} \right \rceil * (PricePerAssignment_\$ + '
            r'max(PricePerAssignment_\$ * TolokaCommission, MinTolokaCommission_\$) * [MinOverlap \dots MaxOverlap] = '
            r'\lceil 50 / 12 \rceil * (0.06\$ + max(0.06\$ * 0.3, 0.001\$) * [2 \dots 4] = '
            r'5 * 0.078 * [2 \dots 4] = 0.78\$ \dots 1.56\$.'
        )


class TestAnnotationPriceFormula:
    def test_simple(self):
        formula = pricing.AnnotationPriceFormula(
            input_objects_count=60,
            markup_config=pricing.PoolPricingConfig(assignment_price=0.06, real_tasks_count=15, control_tasks_count=0),
            check_config=pricing.PoolPricingConfig(assignment_price=0.03, real_tasks_count=20, control_tasks_count=4),
            markup_overlap=classification_loop.DynamicOverlap(min_overlap=3, max_overlap=5, confidence=0.85),
            check_overlap=classification_loop.StaticOverlap(overlap=3),
            assignment_check_sample=None,
        )
        assert (formula.min_total_price_clear, formula.max_total_price_clear) == (approx(1.99, abs=0.01), approx(3.315))

        assert formula.clear_formula() == (
            r'MarkupTolokaComission = max \left( TolokaCommission, \frac {MinTolokaCommission} '
            r'{PricePerMarkupAssignment} \right) = max \left( 0.3, \frac {0.001}{0.06} \right) = 0.3;',
            r'MarkupPrice = MarkupTaskCount * PricePerMarkupTask_\$ * [MinMarkupOverlap \dots MaxMarkupOverlap] * '
            r'(1 + MarkupTolokaComission) = 60 * 0.0040\$ * [3 \dots 5] * 1.3 = 0.94\$ \dots 1.56\$;',
            r'CheckTaskCount = MarkupTaskCount * [MinMarkupOverlap \dots MaxMarkupOverlap] = 60 * [3 \dots 5] = '
            r'[180 \dots 300];',
            r'CheckTolokaComission = max \left( TolokaCommission, \frac {MinTolokaCommission} {PricePerCheckAssignment}'
            r' \right) = max \left( 0.3, \frac {0.001}{0.03} \right) = 0.3;',
            r'CheckPrice = CheckTaskCount * PricePerCheckTask_\$ * CheckOverlap * (1 + CheckTolokaComission) = '
            r'[180 \dots 300] * 0.0015\$ * 3 * 1.3 = 1.05\$ \dots 1.75\$;',
            r'TotalPrice = MarkupPrice + CheckPrice = 0.94\$ \dots 1.56\$ + 1.05\$ \dots 1.75\$ = 1.99\$ \dots 3.31\$.',
        )

    @patch('crowdom.pricing.min_toloka_commission', 0.005)
    def test_complex(self):
        formula = pricing.AnnotationPriceFormula(
            input_objects_count=60,
            markup_config=pricing.PoolPricingConfig(assignment_price=0.01, real_tasks_count=15, control_tasks_count=0),
            check_config=pricing.PoolPricingConfig(assignment_price=0.01, real_tasks_count=20, control_tasks_count=4),
            markup_overlap=classification_loop.DynamicOverlap(min_overlap=3, max_overlap=5, confidence=0.85),
            check_overlap=classification_loop.DynamicOverlap(min_overlap=2, max_overlap=3, confidence=0.85),
            assignment_check_sample=evaluation.AssignmentCheckSample(
                max_tasks_to_check=5, assignment_accuracy_finalization_threshold=0.6
            ),
        )
        assert (formula.min_total_price_clear, formula.max_total_price_clear) == (approx(0.27), approx(0.525))

        assert formula.clear_formula() == (
            r'MarkupTolokaComission = max \left( TolokaCommission, \frac {MinTolokaCommission} '
            r'{PricePerMarkupAssignment} \right) = max \left( 0.3, \frac {0.005}{0.01} \right) = 0.5;',
            r'MarkupPrice = MarkupTaskCount * PricePerMarkupTask_\$ * [MinMarkupOverlap \dots MaxMarkupOverlap] * '
            r'(1 + MarkupTolokaComission) = 60 * 0.0007\$ * [3 \dots 5] * 1.5 = 0.18\$ \dots 0.30\$;',
            r'CheckTaskCount = MarkupTaskCount * \frac {CheckSampleTaskCount} {MarkupTasksPerAssignment} * '
            r'[MinMarkupOverlap \dots MaxMarkupOverlap] = 60 * \frac {5} {15} * [3 \dots 5] = [60 \dots 100];',
            r'CheckTolokaComission = max \left( TolokaCommission, \frac {MinTolokaCommission} {PricePerCheckAssignment}'
            r' \right) = max \left( 0.3, \frac {0.005}{0.01} \right) = 0.5;',
            r'CheckPrice = CheckTaskCount * PricePerCheckTask_\$ * [MinCheckOverlap \dots MaxCheckOverlap] * '
            r'(1 + CheckTolokaComission) = [60 \dots 100] * 0.0005\$ * [2 \dots 3] * 1.5 = 0.09\$ \dots 0.22\$;',
            r'TotalPrice = MarkupPrice + CheckPrice = 0.18\$ \dots 0.30\$ + 0.09\$ \dots 0.22\$ = 0.27\$ \dots 0.52\$.',
        )
